#!/usr/bin/env python
# Copyright (C) 2003-2020  CAMP
# Please see the accompanying LICENSE file for further information.

import os
import re
import sys
from pathlib import Path
from subprocess import PIPE, run
from sysconfig import get_platform

from setuptools import Extension, find_packages, setup
from setuptools.command.build_ext import build_ext as _build_ext
from setuptools.command.develop import develop as _develop
from setuptools.command.install import install as _install

from config import build_interpreter, check_dependencies, write_configuration

assert sys.version_info >= (3, 7)

# Get the current version number:
txt = Path('gpaw/__init__.py').read_text()
version = re.search("__version__ = '(.*)'", txt)[1]
ase_version_required = re.search("__ase_version_required__ = '(.*)'", txt)[1]

description = 'GPAW: DFT and beyond within the projector-augmented wave method'
long_description = Path('README.rst').read_text()

for i, arg in enumerate(sys.argv):
    if arg.startswith('--customize='):
        custom = arg.split('=')[1]
        raise DeprecationWarning(
            f'Please set GPAW_CONFIG={custom} or place {custom} in ' +
            '~/.gpaw/siteconfig.py')

libraries = ['xc']
library_dirs = []
include_dirs = []
extra_link_args = []
extra_compile_args = ['-Wall', '-Wno-unknown-pragmas', '-std=c99']
runtime_library_dirs = []
extra_objects = []
define_macros = [('NPY_NO_DEPRECATED_API', '7'),
                 ('GPAW_NO_UNDERSCORE_CBLACS', '1'),
                 ('GPAW_NO_UNDERSCORE_CSCALAPACK', '1')]
undef_macros = ['NDEBUG']

mpi_libraries = []
mpi_library_dirs = []
mpi_include_dirs = []
mpi_runtime_library_dirs = []
mpi_define_macros = []

parallel_python_interpreter = False
compiler = None
noblas = False
nolibxc = False
fftw = False
scalapack = False
libvdwxc = False
elpa = False

if os.name != 'nt' and run(['which', 'mpicc'], stdout=PIPE).returncode == 0:
    mpicompiler = 'mpicc'
else:
    mpicompiler = None

mpilinker = mpicompiler

# Search and store current git hash if possible
try:
    from ase.utils import search_current_git_hash
    githash = search_current_git_hash('gpaw')
    if githash is not None:
        define_macros += [('GPAW_GITHASH', githash)]
    else:
        print('.git directory not found. GPAW git hash not written.')
except ImportError:
    print('ASE not found. GPAW git hash not written.')

# User provided customizations:
gpaw_config = os.environ.get('GPAW_CONFIG')
if gpaw_config and not Path(gpaw_config).is_file():
    raise FileNotFoundError(gpaw_config)
for siteconfig in [gpaw_config,
                   'siteconfig.py',
                   '~/.gpaw/siteconfig.py']:
    if siteconfig is not None:
        path = Path(siteconfig).expanduser()
        if path.is_file():
            print('Reading configuration from', path)
            exec(path.read_text())
            break
else:  # no break
    if not noblas:
        libraries.append('blas')

if not parallel_python_interpreter and mpicompiler:
    # Build MPI-interface into _gpaw.so:
    compiler = mpicompiler
    define_macros += [('PARALLEL', '1')]

platform_id = os.getenv('CPU_ARCH')
if platform_id:
    os.environ['_PYTHON_HOST_PLATFORM'] = get_platform() + '-' + platform_id

if compiler is not None:
    # A hack to change the used compiler and linker:
    try:
        # distutils is deprecated and will be removed in 3.12
        from distutils.sysconfig import get_config_vars
    except ImportError:
        from sysconfig import get_config_vars

    # If CC is set then the following hack will not work
    assert not os.environ.get('CC'), 'Please unset CC'

    vars = get_config_vars()
    for key in ['CC', 'LDSHARED']:
        if key in vars:
            value = vars[key].split()
            # first argument is the compiler/linker.  Replace with mpicompiler:
            value[0] = compiler
            vars[key] = ' '.join(value)

for flag, name in [(noblas, 'GPAW_WITHOUT_BLAS'),
                   (nolibxc, 'GPAW_WITHOUT_LIBXC'),
                   (fftw, 'GPAW_WITH_FFTW'),
                   (scalapack, 'GPAW_WITH_SL'),
                   (libvdwxc, 'GPAW_WITH_LIBVDWXC'),
                   (elpa, 'GPAW_WITH_ELPA')]:
    if flag:
        define_macros.append((name, '1'))

sources = [Path('c/bmgs/bmgs.c')]
sources += Path('c').glob('*.c')
sources += Path('c/xc').glob('*.c')
if nolibxc:
    for name in ['libxc.c', 'm06l.c',
                 'tpss.c', 'revtpss.c', 'revtpss_c_pbe.c',
                 'xc_mgga.c']:
        sources.remove(Path(f'c/xc/{name}'))
    if 'xc' in libraries:
        libraries.remove('xc')

# Make build process deterministic (for "reproducible build")
sources = [str(source) for source in sources]
sources.sort()

check_dependencies(sources)

# Convert Path objects to str:
library_dirs = [str(dir) for dir in library_dirs]
include_dirs = [str(dir) for dir in include_dirs]

extensions = [Extension('_gpaw',
                        sources,
                        libraries=libraries,
                        library_dirs=library_dirs,
                        include_dirs=include_dirs,
                        define_macros=define_macros,
                        undef_macros=undef_macros,
                        extra_link_args=extra_link_args,
                        extra_compile_args=extra_compile_args,
                        runtime_library_dirs=runtime_library_dirs,
                        extra_objects=extra_objects)]

write_configuration(define_macros, include_dirs, libraries, library_dirs,
                    extra_link_args, extra_compile_args,
                    runtime_library_dirs, extra_objects, mpicompiler,
                    mpi_libraries, mpi_library_dirs, mpi_include_dirs,
                    mpi_runtime_library_dirs, mpi_define_macros)


class build_ext(_build_ext):
    def run(self):
        import numpy as np
        self.include_dirs.append(np.get_include())

        _build_ext.run(self)

        if parallel_python_interpreter:
            include_dirs.append(np.get_include())
            # Also build gpaw-python:
            error = build_interpreter(
                define_macros, include_dirs, libraries,
                library_dirs, extra_link_args, extra_compile_args,
                runtime_library_dirs, extra_objects,
                mpicompiler, mpilinker, mpi_libraries,
                mpi_library_dirs,
                mpi_include_dirs,
                mpi_runtime_library_dirs, mpi_define_macros)
            assert error == 0


def copy_gpaw_python(cmd, dir: str) -> None:
    major, minor = sys.version_info[:2]
    plat = get_platform() + f'-{major}.{minor}'
    source = f'build/bin.{plat}/gpaw-python'
    target = os.path.join(dir, 'gpaw-python')
    cmd.copy_file(source, target)


class install(_install):
    def run(self):
        _install.run(self)
        copy_gpaw_python(self, self.install_scripts)


class develop(_develop):
    def run(self):
        _develop.run(self)
        copy_gpaw_python(self, self.script_dir)


cmdclass = {'build_ext': build_ext}
if parallel_python_interpreter:
    cmdclass['install'] = install
    cmdclass['develop'] = develop

files = ['gpaw-analyse-basis', 'gpaw-basis',
         'gpaw-plot-parallel-timings', 'gpaw-runscript',
         'gpaw-setup', 'gpaw-upfplot']
scripts = [str(Path('tools') / script) for script in files]


setup(name='gpaw',
      version=version,
      description=description,
      long_description=long_description,
      maintainer='GPAW-community',
      maintainer_email='gpaw-users@listserv.fysik.dtu.dk',
      url='https://wiki.fysik.dtu.dk/gpaw',
      license='GPLv3+',
      platforms=['unix'],
      packages=find_packages(),
      entry_points={
          'console_scripts': ['gpaw = gpaw.cli.main:main'],
          'ase.ioformats': ['gpaw-yaml = gpaw.entry_points:gpaw_yaml']},
      setup_requires=['numpy'],
      install_requires=[f'ase>={ase_version_required}',
                        'scipy>=1.2.0',
                        'pyyaml'],
      extras_require={'docs': ['sphinx-rtd-theme',
                               'graphviz'],
                      'devel': ['flake8',
                                'mypy',
                                'pytest-xdist',
                                'interrogate']},
      ext_modules=extensions,
      scripts=scripts,
      cmdclass=cmdclass,
      classifiers=[
          'Development Status :: 6 - Mature',
          'License :: OSI Approved :: '
          'GNU General Public License v3 or later (GPLv3+)',
          'Operating System :: OS Independent',
          'Programming Language :: Python :: 3',
          'Programming Language :: Python :: 3.7',
          'Programming Language :: Python :: 3.8',
          'Programming Language :: Python :: 3.9',
          'Programming Language :: Python :: 3.10',
          'Topic :: Scientific/Engineering :: Physics'])



# web-page: unocc.out
# Start
from gpaw import GPAW

calc = GPAW('gs.gpw')
calc = calc.fixed_density(nbands=42,
                          convergence={'bands': 40},
                          maxiter=1000,
                          txt='unocc.out')
calc.write('unocc.gpw', mode='all')



from ase import Atom, Atoms
from ase.optimize.bfgslinesearch import BFGSLineSearch
from ase.io import read
from gpaw import GPAW

# Relax the Au 55 atoms cluster with both lcao and paw modes
cluster = read('Au55.xyz')
cell = [(17.79365715, 0, 0),
        (0, 19.60846479, 0),
        (0, 0, 19.84025464)]
cluster.set_cell(cell, scale_atoms=False)
cluster.center()

kwargs_lcao = dict(mode='lcao',
                   # basis='dzp',
                   convergence={'density': 0.1, 'energy': 0.1})

calc = GPAW(h=0.18, txt=None, **kwargs_lcao)
cluster.calc = calc

dyn1 = BFGSLineSearch(cluster, trajectory='Au_cluster_lcao.traj')
dyn1.run(fmax=0.02)
e_cluster_lcao = cluster.get_potential_energy()

dyn2 = BFGSLineSearch(cluster, trajectory='Au_cluster_paw.traj')
dyn2.run(fmax=0.02)
e_cluster_paw = cluster.get_potential_energy()

# Relax CO molecule with both lcao and paw modes
CO = Atoms([Atom('C', (1.0, 1.0, 1.0)),
            Atom('O', (1.0, 1.0, 2.3))],
           cell=(12, 12.5, 14.5))
CO.calc = calc
CO.center()
dyn3 = BFGSLineSearch(CO)
dyn3.run(fmax=0.02)
e_CO_lcao = CO.get_potential_energy()

CO.calc = GPAW(h=0.18, tex=None)
dyn4 = BFGSLineSearch(CO)
dyn4.run(fmax=0.02)
e_CO_paw = CO.get_potential_energy()

CO_bond = CO.get_positions()[1][2] - CO.get_positions()[0][2]

# Attach CO molecule onto the Au cluster
pos = []
pos.append(cluster[34].position.copy())
pos.append(cluster[34].position.copy())
pos[0][2] += 1.8
pos[1][2] += 1.8 + CO_bond
CO = Atoms([Atom('C', pos[0]),
            Atom('O', pos[1])])
cluster.extend(CO)

# Relax the CO adsorbed Au cluster with both Lcao and paw modes
cluster.calc = calc
dyn5 = BFGSLineSearch(cluster, trajectory='CO_cluster_lcao.traj')
dyn5.run(fmax=0.02)
e_cocluster_lcao = cluster.get_potential_energy()

cluster.calc = GPAW(h=0.18, tex=None)
dyn6 = BFGSLineSearch(cluster, trajectory='CO_cluster_paw.traj')
dyn6.run(fmax=0.02)
e_cocluster_paw = cluster.get_potential_energy()

# Print results
print('Adsorption energy of CO on Au cluster (lcao):',
      e_cocluster_lcao - e_CO_lcao - e_cluster_lcao)
print('Adsorption energy of CO on Au cluster (paw):',
      e_cocluster_paw - e_CO_paw - e_cluster_paw)



from ase.dft.kpoints import bandpath
from ase.io import read
from ase.parallel import paropen
from gpaw import GPAW

a = read('gs_Bi2Se3.gpw')

G = [0.0, 0.0, 0.0]
L = [0.5, 0.0, 0.0]
F = [0.5, 0.5, 0.0]
Z = [0.5, 0.5, 0.5]
kpts, x, X = bandpath([G, Z, F, G, L], a.cell, npoints=200)

calc = GPAW('gs_Bi2Se3.gpw').fixed_density(kpts=kpts,
                                           symmetry='off',
                                           txt='Bi2Se3_bands.txt')
calc.write('Bi2Se3_bands.gpw')

with paropen('kpath.dat', 'w') as f:
    for k in x:
        print(k, file=f)

with paropen('highsym.dat', 'w') as f:
    for k in X:
        print(k, file=f)



from myqueue.workflow import run


def workflow():
    with run(script='gs.py', cores=8):
        with run(script='td.py', cores=8, tmax='30m'):
            with run(script='tdc.py', cores=8, tmax='30m'):
                run(script='td_replay.py', cores=8, tmax='30m')
                r1 = run(script='spectrum.py')
                r2 = run(script='td_fdm_replay.py')
        with run(script='ksd_init.py'):
            with run(script='fdm_ind.py', deps=[r2]):
                run(script='ind_plot.py')
            run(script='tcm_plot.py', deps=[r1, r2])
        run(script='spec_plot.py', deps=[r1])
        with run(script='td_pulse.py', cores=8, tmax='30m'):
            run(script='plot_pulse.py')




from ase.build import molecule
from gpaw import GPAW
from gpaw import dscf

# Ground state calculation
calc = GPAW(nbands=8,
            h=0.2,
            xc='PBE',
            spinpol=True,
            convergence={'energy': 100,
                         'density': 100,
                         'eigenstates': 1.0e-9,
                         'bands': -1})

CO = molecule('CO')
CO.center(vacuum=3)
CO.calc = calc

E_gs = CO.get_potential_energy()

# Obtain the pseudowavefunctions and projector overlaps of the
# state which is to be occupied. n=5,6 is the 2pix and 2piy orbitals
n = 5
molecule = [0, 1]
wf_u = [kpt.psit_nG[n] for kpt in calc.wfs.kpt_u]
p_uai = [dict([(molecule[a], P_ni[n]) for a, P_ni in kpt.P_ani.items()])
         for kpt in calc.wfs.kpt_u]

# Excited state calculation
calc_es = GPAW(nbands=8,
               h=0.2,
               xc='PBE',
               spinpol=True,
               convergence={'energy': 100,
                            'density': 100,
                            'eigenstates': 1.0e-9,
                            'bands': -1})

CO.calc = calc_es
lumo = dscf.AEOrbital(calc_es, wf_u, p_uai)
# lumo = dscf.MolecularOrbital(calc, weights={0: [0, 0, 0,  1],
#                                             1: [0, 0, 0, -1]})
dscf.dscf_calculation(calc_es, [[1.0, lumo, 1]], CO)

E_es = CO.get_potential_energy()

print('Excitation energy: ', E_es - E_gs)



# web-page: Co_magnons_vs_rc.png
"""Plot the magnon energy as a function of the cutoff radius rc for all the
high-symmetry points of Co (hcp)"""

# General modules
import numpy as np
from matplotlib import rcParams
import matplotlib.pyplot as plt

# Script modules
from gpaw import GPAW
from gpaw.response.heisenberg import calculate_fm_magnon_energies

# ---------- Inputs ---------- #

# Ground state
gpw = 'Co.gpw'

# High-symmetry points
sp_p = [r'$\Gamma$', 'M', 'K', 'A']
spq_pc = [np.array(qc) for qc in
          [[0., 0., 0], [0.5, 0., 0.], [1 / 3., 1 / 3., 0.], [0., 0., 0.5]]]

# Load MFT data
q_qc = np.load('Co_q_qc.npy')
rc_r = np.load('Co_rc_r.npy')
# We study the case, where the two sublattices are equivalent
J_qabr = np.load('Co_Jsph2_qabr.npy')

# Labels and limits
rlabel = r'$r_{\mathrm{c}}\: [\mathrm{\AA}]$'
mwlabel = r'$\hbar\omega$ [meV]'
rlim = (0.4, 1.85)
mwlim = (200., 600.)

filename = 'Co_magnons_vs_rc.png'

# ---------- Script ---------- #

# Extract the magnetization of the unit cell
calc = GPAW(gpw, txt=None)
muc = calc.get_magnetic_moment()

# Calculate the magnon energies
# We distribute the unit cell magnetization evenly between the sites
mm_ar = muc / 2. * np.ones(J_qabr.shape[2:], dtype=float)
E_qnr = calculate_fm_magnon_energies(J_qabr, q_qc, mm_ar)

# We separate the acoustic and optical magnon modes by sorting them
E_qnr = np.sort(E_qnr, axis=1)

# Make a subplot for each magnon mode
fig, axes = plt.subplots(1, 2, constrained_layout=True)
colors = rcParams['axes.prop_cycle'].by_key()['color']

# Plot the magnon energies at the high-symmetry points as a function of rc
for p, (sp, spq_c) in enumerate(zip(sp_p, spq_pc)):
    q = 0
    while not np.allclose(q_qc[q], spq_c):
        q += 1
        if q == len(q_qc):
            raise ValueError
    for n in range(2):
        if n == 0 and p == 0:
            continue  # Do not plot the acoustic mode Gamma point
        E_r = E_qnr[q, n] * 1.e3  # eV -> meV
        axes[n].plot(rc_r, E_r, '-x',
                     color=colors[p], label=sp)

for n, (ax, mode) in enumerate(zip(axes, ['Acoustic', 'Optical'])):
    ax.set_title(mode)
    ax.set_xlabel(rlabel)
    ax.set_ylabel(mwlabel)
    ax.set_xlim(rlim)
    ax.set_ylim(mwlim)

    ax.legend()

plt.savefig(filename, format=filename.split('.')[-1],
            bbox_inches='tight')



from gpaw import restart

# J.Phys.: Condens. Matter 18 (2006) 41-54
pw91vasp = {'NO': 0.95,
            'N2': 0.00,
            'O2': 0.00,
            'NRu001': -0.94,
            'ORu001': -2.67,
            'Ru001': 0.00}

pbe = {}
pw91 = {}
for name in ['NO', 'O2', 'N2', 'Ru001', 'NRu001', 'ORu001']:
    a, calc = restart(name + '.gpw', txt=None)
    pbe[name] = a.get_potential_energy()
    pw91[name] = pbe[name] + calc.get_xc_difference('PW91')

energies = {}
for data, text in [(pbe, 'PBE'),
                   (pw91, 'PW91 (non-selfconsitent)'),
                   (pw91vasp, 'PW91 (VASP)')]:
    E = [data['NRu001'] - data['Ru001'] - data['N2'] / 2,
         data['ORu001'] - data['Ru001'] - data['O2'] / 2,
         data['NO'] - data['N2'] / 2 - data['O2'] / 2]
    print('{:24} {:.3f} {:.3f} {:.3f}'.format(text, *E))
    energies[text] = E

for e1, e2 in zip(energies['PW91 (non-selfconsitent)'],
                  energies['PW91 (VASP)']):
    assert abs(e1 - e2) < 0.1



from ase.build import bulk

from gpaw import GPAW, FermiDirac
from gpaw.elph.electronphonon import ElectronPhononCoupling

a = 3.567
atoms = bulk('C', 'diamond', a=a)

calc = GPAW(mode='lcao',
            basis='dzp',
            kpts=(5, 5, 5),
            xc='PBE',
            occupations=FermiDirac(0.01),
            symmetry={'point_group': False},
            convergence={'bands': 'nao'},
            txt='elph.txt')

atoms.calc = calc
atoms.get_potential_energy()
calc.write("scf.gpw", 'all')

elph = ElectronPhononCoupling(atoms, atoms.calc, calculate_forces=False)
elph.set_lcao_calculator(atoms.calc)
elph.run()
elph.calculate_supercell_matrix()



import subprocess
import pytest
from ase.io import write
from ase.units import Bohr
from gpaw import GPAW
from gpaw.utilities.ps2ae import PS2AE
from gpaw.utilities.bader import read_bader_charges


@pytest.mark.serial
def test_bader(gpw_files, in_tmp_dir):
    """Test bader analysis on interpolated density."""
    calc = GPAW(gpw_files['c2h4_pw_nosym'])
    converter = PS2AE(calc)
    density = converter.get_pseudo_density()
    ne = density.sum() * converter.dv
    assert ne == pytest.approx(12, abs=1e-5)

    write('density.cube', calc.atoms, data=density * Bohr**3)
    try:
        subprocess.run('bader -p all_atom density.cube'.split())
    except FileNotFoundError:
        return

    charges = read_bader_charges()
    assert charges[:2].ptp() == 0.0
    assert charges[2:].ptp() < 0.001
    assert charges.sum() == pytest.approx(12.0, abs=0.0001)



import numpy as np
import pytest
from gpaw.core import UniformGrid, PlaneWaves
from gpaw.mpi import world
from gpaw.core.plane_waves import find_reciprocal_vectors
from math import pi


@pytest.mark.ci
def test_pw_redist():
    a = 2.5
    pw = PlaneWaves(ecut=10, cell=[a, a, a], comm=world)
    f1 = pw.empty()
    f1.data[:] = 1.0
    f2 = f1.gather()
    if f2 is not None:
        assert (f2.data == 1.0).all()
        assert f2.desc.comm.size == 1


@pytest.mark.ci
def test_pw_map():
    """Test mapping from 5 to 9 G-vectors."""
    pw5 = PlaneWaves(ecut=0.51 * (2 * pi)**2,
                     cell=[1, 1, 0.1],
                     dtype=complex)
    pw9 = PlaneWaves(ecut=1.01 * (2 * pi)**2,
                     cell=[1, 1, 0.1],
                     dtype=complex,
                     comm=world)
    my_G_g, g_r = pw9.map_indices(pw5)
    print(world.rank, my_G_g, g_r)
    expected = {
        1: ([[0, 1, 2, 3, 6]],
            [[0, 1, 2, 3, 4]]),
        2: ([[0, 1, 2, 3], [1]],
            [[0, 1, 2, 3], [4]]),
        4: ([[0, 1, 2], [0], [0], []],
            [[0, 1, 2], [3], [4], []]),
        8: ([[0, 1], [0, 1], [], [0], [], [], [], []],
            [[0, 1], [2, 3], [], [4], [], [], [], []])}
    assert not (my_G_g - expected[world.size][0][world.rank]).any()
    for g, g0 in zip(g_r, expected[world.size][1]):
        assert not (np.array(g) - g0).any()


def test_pw_integrate():
    a = 1.0
    decomp = {1: [[0, 4], [0, 4], [0, 4]],
              2: [[0, 2, 4], [0, 4], [0, 4]],
              4: [[0, 2, 4], [0, 2, 4], [0, 4]],
              8: [[0, 1, 2, 3, 4], [0, 2, 4], [0, 4]]}[world.size]
    grid = UniformGrid(cell=[a, a, a], size=(4, 4, 4), comm=world,
                       decomp=decomp)
    gridc = grid.new(dtype=complex)

    g1 = grid.empty()
    g1.data[:] = 1.0

    g2 = grid.empty()
    g2.data[:] = 1.0
    g2.data += [0, 1, 0, -1]

    g3 = gridc.empty()
    g3.data[:] = 1.0

    g4 = gridc.empty()
    g4.data[:] = 1.0
    g4.data += [0, 1, 0, -1]

    g5 = gridc.empty()
    g5.data[:] = 1.0 + 1.0j
    g5.data += [0, 1, 0, -1]

    ecut = 0.5 * (2 * np.pi / a)**2 * 1.01
    for g in [g1, g2, g3, g4, g5]:
        pw = PlaneWaves(cell=g.desc.cell, dtype=g.desc.dtype,
                        ecut=ecut, comm=world)
        f = g.fft(pw=pw)
        print(f.data)

        gg = g.new()
        gg.scatter_from(f.gather(broadcast=True)
                        .ifft(grid=g.desc.new(comm=None)))
        assert (g.data == gg.data).all()

        i1 = g.integrate()
        i2 = f.integrate()
        assert i1 == i2
        assert type(i1) == g.desc.dtype

        i1 = g.integrate(g)
        i2 = f.integrate(f)
        assert i1 == i2
        assert type(i1) == g.desc.dtype

        g1 = g.desc.empty(1)
        g1.data[:] = g.data
        m1 = g1.matrix_elements(g1)
        assert (i1 == m1.data).all()

        f1 = f.desc.empty(1)
        f1.data[:] = f.data
        m2 = f1.matrix_elements(f1)
        assert (i2 == m2.data).all()


def test_grr():
    from ase.units import Ha, Bohr
    grid = UniformGrid(cell=[2 / Bohr, 2 / Bohr, 2.737166 / Bohr],
                       size=(9, 9, 12),
                       comm=world)
    pw = PlaneWaves(ecut=340 / Ha, cell=grid.cell, comm=world)
    print(pw.G_plus_k_Gv.shape)
    from gpaw.grid_descriptor import GridDescriptor
    from gpaw.pw.descriptor import PWDescriptor
    g = GridDescriptor((9, 9, 12), [2 / Bohr, 2 / Bohr, 2.737166 / Bohr])
    p = PWDescriptor(340 / Ha, g)
    print(p.get_reciprocal_vectors().shape)
    assert (p.get_reciprocal_vectors() == pw.G_plus_k_Gv).all()


def test_find_g():
    G, e, i = find_reciprocal_vectors(0.501 * (2 * pi)**2,
                                      np.eye(3))
    assert i.T.tolist() == [[0, 0, 0],
                            [0, 0, 1],
                            [0, 0, -1],
                            [0, 1, 0],
                            [0, -1, 0],
                            [1, 0, 0],
                            [-1, 0, 0]]
    G, e, i = find_reciprocal_vectors(0.501 * (2 * pi)**2,
                                      np.eye(3),
                                      dtype=float)
    assert i.T.tolist() == [[0, 0, 0],
                            [0, 0, 1],
                            [0, 1, 0],
                            [1, 0, 0]]
    G, e, i = find_reciprocal_vectors(0.5 * (2 * pi)**2,
                                      np.eye(3),
                                      kpt=np.array([0.1, 0, 0]))
    assert i.T.tolist() == [[0, 0, 0],
                            [-1, 0, 0]]



from ase.build import bulk
from ase.phonons import Phonons

from gpaw import GPAW, FermiDirac

a = 3.567
atoms = bulk('C', 'diamond', a=a)

calc = GPAW(mode='lcao',
            basis='dzp',
            kpts=(5, 5, 5),
            xc='PBE',
            occupations=FermiDirac(0.01),
            symmetry={'point_group': False},
            convergence={'energy': 2e-5, 'density': 1e-5},
            txt='phonons.txt')

atoms.calc = calc

# Phonon calculator
ph = Phonons(atoms, atoms.calc, supercell=(1, 1, 1), delta=0.01)
ph.run()



from pathlib import Path
from gpaw.mpi import world
from gpaw.response.df import DielectricFunction
from gpaw.response.qeh import BuildingBlock

df = DielectricFunction(calc='WSe2_gs_fulldiag.gpw',
                        eta=0.001,
                        frequencies={'type': 'nonlinear',
                                     'domega0': 0.05,
                                     'omega2': 10.0},
                        nblocks=8,
                        ecut=150,
                        truncation='2D')

buildingblock = BuildingBlock('WSe2', df, qmax=3.0)

buildingblock.calculate_building_block()

if world.rank == 0:
    Path('WSe2_gs_fulldiag.gpw').unlink()



"""
Python wrapper for FFTW3 library
================================

.. autoclass:: FFTPlans

"""
from __future__ import annotations

import numpy as np
from scipy.fft import fftn, ifftn, irfftn, rfftn

import _gpaw
from gpaw.typing import Array3D, DTypeLike, IntVector

ESTIMATE = 64
MEASURE = 0
PATIENT = 32
EXHAUSTIVE = 8


def have_fftw() -> bool:
    """Did we compile with FFTW?"""
    return hasattr(_gpaw, 'FFTWPlan')


def check_fft_size(n: int, factors=[2, 3, 5, 7]) -> bool:
    """Check if n is an efficient fft size.

    Efficient means that n can be factored into small primes (2, 3, 5, 7).

    >>> check_fft_size(17)
    False
    >>> check_fft_size(18)
    True
    """

    if n == 1:
        return True
    for x in factors:
        if n % x == 0:
            return check_fft_size(n // x, factors)
    return False


def get_efficient_fft_size(N: int, n=1, factors=[2, 3, 5, 7]) -> int:
    """Return smallest efficient fft size.

    Must be greater than or equal to N and divisible by n.

    >>> get_efficient_fft_size(17)
    18
    """
    N = -(-N // n) * n
    while not check_fft_size(N, factors):
        N += n
    return N


def empty(shape, dtype=float):
    """numpy.empty() equivalent with 16 byte alignment."""
    assert dtype == complex
    N = np.prod(shape)
    a = np.empty(2 * N + 1)
    offset = (a.ctypes.data % 16) // 8
    a = a[offset:2 * N + offset].view(complex)
    a.shape = shape
    return a


def create_plans(size_c: IntVector,
                 dtype: DTypeLike,
                 flags: int = MEASURE) -> FFTPlans:
    """Create plan-objects for FFT and inverse FFT."""
    if have_fftw():
        return FFTWPlans(size_c, dtype, flags)
    return NumpyFFTPlans(size_c, dtype)


class FFTPlans:
    def __init__(self,
                 size_c: IntVector,
                 dtype: DTypeLike):
        if dtype == float:
            rsize_c = (size_c[0], size_c[1], size_c[2] // 2 + 1)
            self.tmp_Q = empty(rsize_c, complex)
            self.tmp_R = self.tmp_Q.view(float)[:, :, :size_c[2]]
        else:
            self.tmp_Q = empty(size_c, complex)
            self.tmp_R = self.tmp_Q

    def fft(self) -> None:
        """Do FFT from ``tmp_R`` to ``tmp_Q``.

        >>> plans = create_plans([4, 1, 1], float)
        >>> plans.tmp_R[:, 0, 0] = [1, 0, 1, 0]
        >>> plans.fft()
        >>> plans.tmp_Q[:, 0, 0]
        array([2.+0.j, 0.+0.j, 2.+0.j, 0.+0.j])
        """
        raise NotImplementedError

    def ifft(self) -> None:
        """Do inverse FFT from ``tmp_Q`` to ``tmp_R``.

        >>> plans = create_plans([4, 1, 1], complex)
        >>> plans.tmp_Q[:, 0, 0] = [0, 1j, 0, 0]
        >>> plans.ifft()
        >>> plans.tmp_R[:, 0, 0]
        array([ 0.+1.j, -1.+0.j,  0.-1.j,  1.+0.j])
        """
        raise NotImplementedError


class FFTWPlans(FFTPlans):
    """FFTW3 3d transforms."""
    def __init__(self, size_c, dtype, flags=MEASURE):
        if not have_fftw():
            raise ImportError('Not compiled with FFTW.')
        super().__init__(size_c, dtype)
        self._fftplan = _gpaw.FFTWPlan(self.tmp_R, self.tmp_Q, -1, flags)
        self._ifftplan = _gpaw.FFTWPlan(self.tmp_Q, self.tmp_R, 1, flags)

    def fft(self):
        _gpaw.FFTWExecute(self._fftplan)

    def ifft(self):
        _gpaw.FFTWExecute(self._ifftplan)

    def __del__(self):
        _gpaw.FFTWDestroy(self._fftplan)
        _gpaw.FFTWDestroy(self._ifftplan)


class NumpyFFTPlans(FFTPlans):
    """Numpy fallback."""
    def fft(self):
        if self.tmp_R.dtype == float:
            self.tmp_Q[:] = rfftn(self.tmp_R, overwrite_x=True)
        else:
            self.tmp_Q[:] = fftn(self.tmp_R, overwrite_x=True)

    def ifft(self):
        if self.tmp_R.dtype == float:
            self.tmp_R[:] = irfftn(self.tmp_Q, self.tmp_R.shape,
                                   norm='forward', overwrite_x=True)
        else:
            self.tmp_R[:] = ifftn(self.tmp_Q, self.tmp_R.shape,
                                  norm='forward', overwrite_x=True)


# The rest of this file will be removed in the future ...

def check_fftw_inputs(in_R, out_R):
    for arr in in_R, out_R:
        # Note: Arrays not necessarily contiguous due to 16-byte alignment
        assert arr.ndim == 3  # We can perhaps relax this requirement
        assert arr.dtype == float or arr.dtype == complex

    if in_R.dtype == out_R.dtype == complex:
        assert in_R.shape == out_R.shape
    else:
        # One real and one complex:
        R, C = (in_R, out_R) if in_R.dtype == float else (out_R, in_R)
        assert C.dtype == complex
        assert R.shape[:2] == C.shape[:2]
        assert C.shape[2] == 1 + R.shape[2] // 2


class FFTPlan:
    """FFT 3d transform."""
    def __init__(self,
                 in_R: Array3D,
                 out_R: Array3D,
                 sign: int,
                 flags: int = MEASURE):
        check_fftw_inputs(in_R, out_R)
        self.in_R = in_R
        self.out_R = out_R
        self.sign = sign
        self.flags = flags

    def execute(self) -> None:
        raise NotImplementedError


class FFTWPlan(FFTPlan):
    """FFTW3 3d transform."""
    def __init__(self, in_R, out_R, sign, flags=MEASURE):
        if not have_fftw():
            raise ImportError('Not compiled with FFTW.')
        self._ptr = _gpaw.FFTWPlan(in_R, out_R, sign, flags)
        FFTPlan.__init__(self, in_R, out_R, sign, flags)

    def execute(self):
        _gpaw.FFTWExecute(self._ptr)

    def __del__(self):
        if getattr(self, '_ptr', None) and _gpaw is not None:
            _gpaw.FFTWDestroy(self._ptr)
        self._ptr = None


class NumpyFFTPlan(FFTPlan):
    """Numpy fallback."""
    def execute(self):
        if self.in_R.dtype == float:
            self.out_R[:] = np.fft.rfftn(self.in_R)
        elif self.out_R.dtype == float:
            self.out_R[:] = np.fft.irfftn(self.in_R, self.out_R.shape)
            self.out_R *= self.out_R.size
        elif self.sign == 1:
            self.out_R[:] = np.fft.ifftn(self.in_R, self.out_R.shape)
            self.out_R *= self.out_R.size
        else:
            self.out_R[:] = np.fft.fftn(self.in_R)


def create_plan(in_R: Array3D,
                out_R: Array3D,
                sign: int,
                flags: int = MEASURE) -> FFTPlan:
    if have_fftw():
        return FFTWPlan(in_R, out_R, sign, flags)
    return NumpyFFTPlan(in_R, out_R, sign, flags)



import io
from itertools import chain

import matplotlib.pyplot as plt
import numpy as np
from ase.io import read
from ase.optimize import FIRE
from ase.vibrations import Vibrations
from ase.vibrations.franck_condon import FranckCondon

from gpaw import GPAW
from gpaw.mpi import world
from gpaw.cluster import Cluster
from gpaw.utilities.folder import Folder

butadiene = u"""10

C       3.649801161546418      5.442281389577507      3.863313703750026
C       5.051651240044169      5.368220758269772      4.162165876906096
C       5.750174626862403      4.162261915959347      4.240449977068684
C       7.150130182125531      4.155384186721486      4.537328602062397
H       3.218154657585170      4.565210696328925      3.522601038049320
H       3.077656122062729      6.375092902842770      3.826039498180272
H       5.478464901706067      6.370680001794822      4.422235395756437
H       5.320549047980879      3.220584852467720      3.974551561510350
H       7.723359150977955      3.224855971783890      4.574146712279462
H       7.580803493981530      5.034479218283977      4.877211530909463
"""

h = 0.3
atoms = Cluster(read(io.StringIO(butadiene), format='xyz'))
atoms.minimal_box(3.0, h)
atoms.calc = GPAW(h=h)
if 0:
    dyn = FIRE(atoms)
    dyn.run(fmax=0.05)
    atoms.write('butadiene.xyz')

vibname = 'fcvib'
vib = Vibrations(atoms, name=vibname)
vib.run()

# Modul
a = FranckCondon(atoms, vibname, minfreq=250)

# excited state forces
F = np.array([[-2.11413, 0.07317, -0.91682],
              [3.23569, -0.74520, 0.76758],
              [-3.44847, 0.63846, -0.81080],
              [2.77345, 0.01272, 0.74811],
              [-0.06544, -0.01078, -0.03209],
              [-0.01245, -0.01123, -0.00040],
              [0.00186, -0.05864, -0.00371],
              [-0.00151, 0.05815, 0.00141],
              [0.01625, 0.00781, -0.00202],
              [0.06253, 0.00902, 0.03381]])

# Huang-Rhys factors
S, fq = a.get_Huang_Rhys_factors(F)

# Temperature and #quanta taken into account
T = 300
n = 10

colors = ['red', 'blue', 'green']
labels = ['normal modes with #quanta=1,-1 and the 0-0 transition',
          'normal modes with #quanta>1,<-1',
          'combination of two normal modes']

S = np.append(1, S)
fq = np.append(0, fq)
if world.rank == 0:
    plt.vlines(fq, 0, S, color='darkblue',
               label='Huang-Rhys factors', linewidth=2.0)
    
    plt.legend(loc='upper right')
    plt.axis([-1000, 6000, 0, 1.05])
    plt.ylabel('HR factors [a.u.]')
    plt.xlabel('frequency [cm-1]')
    plt.show()

# Plot Franck-Condon factors
FC, f = a.get_Franck_Condon_factors(T, F, order=n)
for l in range(len(FC)):
    plt.vlines(f[l], 0, FC[l], color=colors[l], label=labels[l], linewidth=2.0)

# Fold the spectrum with a gaussian function
width = 300
folding = 'Gauss'
x = [j for j in chain(*f)]
y = [j for j in chain(*FC)]

if world.rank == 0:
    X, Y = Folder(width, folding).fold(x, y)
    plt.plot(X, Y * 750, color='black', linewidth=2.0,
             label='Temperature=' + str(T) + 'K, #quanta=' + str(n) +
             ', width of gaussian=' + str(width) + 'cm^-1',
             linestyle='dashed')

    plt.legend(loc='upper right')
    plt.axis([-1000, 6000, 0, 0.6])
    plt.ylabel('FC factors [a.u.]')
    plt.xlabel('frequency [cm-1]')
    plt.show()



import pytest

from ase import Atoms
from gpaw import GPAW, LCAO
from gpaw.directmin.derivatives import Derivatives
import numpy as np


def test_directmin_lcao_numerical_hessian(in_tmp_dir):
    """
    Test complex numerical Hessian
    w.r.t rotation parameters in LCAO

    :param in_tmp_dir:
    :return:
    """

    calc = GPAW(xc='PBE',
                mode=LCAO(force_complex_dtype=True),
                h=0.25,
                basis='dz(dzp)',
                spinpol=False,
                eigensolver={'name': 'etdm',
                             'representation': 'u-invar'},
                occupations={'name': 'fixed-uniform'},
                mixer={'backend': 'no-mixing'},
                nbands='nao',
                symmetry='off',
                )

    atoms = Atoms('H', positions=[[0, 0, 0]])
    atoms.center(vacuum=5.0)
    atoms.set_pbc(False)
    atoms.calc = calc
    atoms.get_potential_energy()

    numder = Derivatives(calc.wfs.eigensolver, calc.wfs)

    hess_n = numder.get_numerical_derivatives(
        calc.wfs.eigensolver,
        calc.hamiltonian,
        calc.wfs,
        calc.density,
        what2calc='hessian'
    )
    hess_a = numder.get_analytical_derivatives(
        calc.wfs.eigensolver,
        calc.hamiltonian,
        calc.wfs,
        calc.density,
        what2calc='hessian'
    )
    hess_nt = np.asarray([[1.32720630e+00, -1.93947467e-11],
                         [3.95786680e-09, 1.14599176e+00]])
    assert hess_n == pytest.approx(hess_nt, abs=1e-4)
    assert hess_a == pytest.approx(hess_nt, abs=0.2)

    a_mat_u = {0: [np.sqrt(2) * np.pi / 4.0 + 1.0j * np.sqrt(2) * np.pi / 4.0]}
    c_nm_ref = calc.wfs.eigensolver.dm_helper.reference_orbitals
    calc.wfs.eigensolver.rotate_wavefunctions(calc.wfs,
                                              a_mat_u,
                                              {0: calc.wfs.bd.nbands},
                                              c_nm_ref
                                              )
    calc.wfs.eigensolver.update_ks_energy(calc.hamiltonian,
                                          calc.wfs, calc.density)
    calc.wfs.eigensolver.get_canonical_representation(
        calc.hamiltonian, calc.wfs, calc.density, sort_eigenvalues=True)
    c_nm = {x: calc.wfs.kpt_u[x].C_nM.copy()
            for x in range(len(calc.wfs.kpt_u))}

    numder = Derivatives(calc.wfs.eigensolver, calc.wfs,
                         c_ref=c_nm)

    hess_n = numder.get_numerical_derivatives(
        calc.wfs.eigensolver,
        calc.hamiltonian,
        calc.wfs,
        calc.density,
        what2calc='hessian'
    )

    hess_a = numder.get_analytical_derivatives(
        calc.wfs.eigensolver,
        calc.hamiltonian,
        calc.wfs,
        calc.density,
        what2calc='hessian'
    )
    hess_nt = np.asarray([[-1.08209601e+00, -1.11022302e-09],
                         [8.50014503e-10, -9.37664521e-01]])
    assert hess_n == pytest.approx(hess_nt, abs=1e-4)
    assert hess_a == pytest.approx(hess_nt, abs=0.2)



import numpy as np
import pytest
from gpaw.core import PlaneWaves, UniformGrid
from gpaw.mpi import world


@pytest.mark.ci
def test_acf():
    a = 2.5
    n = 20

    # comm = world.new_communicator([world.rank])
    grid = UniformGrid(cell=[a, a, a], size=(n, n, n), comm=world,
                       dtype=complex)

    alpha = 4.0
    s = (0, 3.0, lambda r: np.exp(-alpha * r**2))

    basis = grid.atom_centered_functions(
        [[s]],
        positions=[[0.5, 0.5, 0.5]])
    coefs = basis.layout.empty()
    if 0 in coefs:
        print(coefs[0])
        coefs[0] = [1.0]
    f1 = grid.zeros()
    basis.add_to(f1, coefs)

    if 0:
        from sympy import integrate, exp, oo, var
        a, r = var('a, r')
        integrate(exp(-a * r**2) * r**2, (r, 0, oo))

    I = np.pi / 2 / alpha**1.5
    assert f1.integrate() == pytest.approx(I)

    f2 = f1.gather(broadcast=True)
    x, y = f2.xy(10, 10, ...)
    y0 = np.exp(-alpha * (x - a / 2)**2) / (4 * np.pi)**0.5
    assert abs(y - y0).max() == pytest.approx(0.0, abs=0.001)

    pw = PlaneWaves(ecut=50, cell=grid.cell, dtype=complex, comm=world)

    basis = pw.atom_centered_functions(
        [[s]],
        positions=[[0.5, 0.5, 0.5]])

    coefs = basis.layout.empty()
    if 0 in coefs:
        print(coefs[0])
        coefs[0] = [1.0]

    f1 = pw.zeros()
    basis.add_to(f1, coefs)
    assert f1.integrate() == pytest.approx(I)
    f2 = f1.gather(broadcast=True)
    r2 = f2.ifft(grid=grid.new(comm=None))
    x, y = r2.xy(10, 10, ...)
    assert abs(y - y0).max() == pytest.approx(0.0, abs=0.002)



# web-page: fig2.png
import numpy as np
import matplotlib.pyplot as plt
from ase.units import Bohr
from gpaw import GPAW
rs = 5.0 * Bohr
calc = GPAW('surface.gpw', txt=None)
density = calc.get_pseudo_density()[0, 0]
h = 0.2
a = 8 * h
v = 3 * a
L = 10 * a
z = np.linspace(0, v + L + v, len(density), endpoint=False)
# Position of surface is between two grid points:
z0 = (v + L - h / 2)
n = 1 / (4 * np.pi / 3 * rs**3)  # electron density
kF = (3 * np.pi**2 * n)**(1.0 / 3)
lambdaF = 2 * np.pi / kF  # Fermi wavelength
plt.figure(figsize=(6, 6 / 2**0.5))
plt.plot([-L / lambdaF, -L / lambdaF, 0, 0], [0, 1, 1, 0], 'k')
plt.plot((z - z0) / lambdaF, density / n)
# plt.xlim(xmin=-1.2, xmax=1)
plt.ylim(ymin=0)
plt.title(r'$r_s=%.1f\ \mathrm{Bohr},\ \lambda_F=%.1f\ \mathrm{Bohr}$' %
          (rs / Bohr, lambdaF / Bohr))
plt.xlabel('DISTANCE (FERMI WAVELENGTHS)')
plt.ylabel('ELECTRON DENSITY')
plt.savefig('fig2.png')



# web-page: xas_h2o_spectrum.png, h2o_xas_box.png

def check():
    exec(open('plot.py').read())
    e_dks = float(open('dks.result').readline().split()[2])
    assert abs(e_dks - 532.502) < 0.001
    exec(open('h2o_xas_box2.py').read())


def workflow():
    from myqueue.workflow import run
    with run(script='setups.py'):
        r1 = run(script='run.py', cores=8, tmax='25m')
        r2 = run(script='dks.py', cores=8, tmax='25m')
        r3 = run(script='h2o_xas_box1.py', cores=8, tmax='25m')
    run(function=check, deps=[r1, r2, r3])



from myqueue.workflow import run


def workflow():
    with run(script='gs_N2.py', cores=16):
        r1 = run(script='frequency.py', tmax='3h')
        r2 = run(script='con_freq.py', cores=2, tmax='16h')
        r3 = run(script='rpa_N2.py', cores=48, tmax='1h')
    with r1, r2:
        run(script='plot_w.py')
    with r2:
        run(script='plot_con_freq.py')
    with r3:
        run(script='extrapolate.py')

    with r1, r2, r3:
        run(shell='rm', args=['N.gpw', 'N2.gpw'])  # clean up

    with run(script='gs_graph_Co.py', cores=16, tmax='1d'):
        with run(script='rpa_graph_Co.py', cores=16, tmax='10h'):
            run(shell='rm', args=['gs_*.gpw'])  # clean up



import numpy as np
from ase.units import Ha, Bohr


class ResponseGroundStateAdapter:
    def __init__(self, calc):
        wfs = calc.wfs

        self.kd = wfs.kd
        self.world = calc.world
        self.gd = wfs.gd
        self.bd = wfs.bd
        self.nspins = wfs.nspins
        self.dtype = wfs.dtype

        self.spos_ac = calc.spos_ac

        self.kpt_u = wfs.kpt_u
        self.kpt_qs = wfs.kpt_qs
        self.setups = wfs.setups

        self.fermi_level = wfs.fermi_level
        self.atoms = calc.atoms
        self.pbc = self.atoms.pbc
        self.volume = self.gd.volume

        self.nvalence = wfs.nvalence

        self._wfs = wfs
        self._density = calc.density
        self._hamiltonian = calc.hamiltonian
        self._calc = calc

    @property
    def pd(self):
        # This is an attribute error in FD/LCAO mode.
        # We need to abstract away "calc" in all places used by response
        # code, and that includes places that are also compatible with FD.
        return self._wfs.pd

    def get_occupations_width(self):
        # Ugly hack only used by pair.intraband_pair_density I think.
        # Actually: was copy-pasted in chi0 also.
        # More duplication can probably be eliminated around those.

        # Only works with Fermi-Dirac distribution
        occs = self._wfs.occupations
        assert occs.name in {'fermi-dirac', 'zero-width'}

        # No carriers when T=0
        width = getattr(occs, '_width', 0.0) / Ha
        return width

    def nonpbc_cell_product(self):
        """Volume, area, or length, taken in all non-periodic directions."""
        nonpbc = ~self.pbc
        cell_cv = self.gd.cell_cv
        return abs(np.linalg.det(cell_cv[nonpbc][:, nonpbc]))

    @property
    def nt_sG(self):
        # Used by kxc
        return self._density.nt_sG

    @property
    def D_asp(self):
        # Used by kxc
        return self._density.D_asp

    def all_electron_density(self, gridrefinement=2):
        # used by kxc
        return self._density.get_all_electron_density(
            atoms=self.atoms, gridrefinement=gridrefinement)

    # Things used by EXX.  This is getting pretty involved.
    #
    # EXX naughtily accesses the density object in order to
    # interpolate_pseudo_density() which is in principle mutable.

    def hacky_all_electron_density(self, **kwargs):
        # fxc likes to get all electron densities.  It calls
        # calc.get_all_electron_density() and so we wrap that here.
        # But it also collects to serial (bad), and it also zeropads
        # nonperiodic directions (probably WRONG!).
        #
        # Also this one returns in user units, whereas the calling
        # code actually wants internal units.  Very silly then.
        #
        # ALso, the calling code often wants the gd, which is not
        # returned, so it is redundantly reconstructed in multiple
        # places by refining the "right" number of times.
        n_g = self._calc.get_all_electron_density(**kwargs)
        n_g *= Bohr**3
        return n_g

    # Used by EXX.
    @property
    def hamiltonian(self):
        return self._hamiltonian

    # Used by EXX.
    @property
    def density(self):
        return self._density

    # Ugh SOC
    def soc_eigenstates(self, **kwargs):
        from gpaw.spinorbit import soc_eigenstates
        return soc_eigenstates(self._calc, **kwargs)

    @property
    def xcname(self):
        return self.hamiltonian.xc.name

    # XXX This is used by xc == JGMsx from g0w0
    def get_band_gap(self):
        from ase.dft.bandgap import get_band_gap
        gap, k1, k2 = get_band_gap(self._calc)
        return gap

    def get_xc_difference(self, xc):
        # XXX used by gpaw/xc/tools.py
        return self._calc.get_xc_difference(xc)

    def get_wave_function_array(self, u, n):
        # XXX used by gpaw/xc/tools.py in a hacky way
        return self._wfs._get_wave_function_array(
            u, n, realspace=True)



from pathlib import Path

from ase.io import read
from myqueue.workflow import run


def workflow():
    with run(script='domom_co.py', cores=8):
        run(function=check_co)
    with run(script='mom_h2o.py', cores=8):
        run(function=check_h2o)


def check_co():
    for tag in ['spinpol', 'spinpaired']:
        co = read('co_' + tag + '.txt')
        assert abs(co.get_distance(0, 1) - 1.248) < 0.01


def check_h2o():
    text = Path('h2o_energies.txt').read_text()
    for line in text.splitlines():
        if line.startswith('Excitation energy triplet'):
            et = float(line.split()[-2])
        elif line.startswith('Excitation energy singlet'):
            es = float(line.split()[-2])
    assert abs(et - 9.21) < 0.005
    assert abs(es - 9.68) < 0.005



import pytest
import numpy as np

from ase import Atoms

from gpaw import GPAW
from gpaw.kpt_descriptor import KPointDescriptor
from gpaw.pw.descriptor import PWDescriptor
from gpaw.response.pair import PairDensity
from gpaw.response.math_func import two_phi_nabla_planewave_integrals


@pytest.mark.response
def test_response_pair(in_tmp_dir, scalapack):
    nb = 6

    a = Atoms('H', cell=[2.5] * 3, pbc=True)

    calc = GPAW(mode='pw', kpts=[[0, 0, 0], [0.25, 0, 0]],
                parallel=dict(domain=1))
    a.calc = calc
    a.get_potential_energy()
    calc.diagonalize_full_hamiltonian(nbands=nb, expert=True)
    calc.write('a.gpw', 'all')

    pair = PairDensity('a.gpw')

    # Check continuity eq.
    for q_c in [[0, 0, 0], [1. / 4, 0, 0]]:
        ol = np.allclose(q_c, 0.0)
        qd = KPointDescriptor([q_c])
        pd = PWDescriptor(0.25, calc.wfs.gd, complex, qd)
        kptpair = pair.get_kpoint_pair(pd, 0, [0, 0, 0],
                                       0, nb, 0, nb)
        deps_nm = kptpair.get_transition_energies(np.arange(0, nb),
                                                  np.arange(0, nb))

        n_nmG = pair.get_pair_density(pd, kptpair, np.arange(0, nb),
                                      np.arange(0, nb), optical_limit=ol)

        n_nmvG = pair.get_pair_momentum(pd, kptpair, np.arange(0, nb),
                                        np.arange(0, nb))

        if ol:
            n2_nmv = np.zeros((nb, nb, 3), complex)
            for n in range(0, nb):
                n2_nmv[n] = pair.optical_pair_velocity(n, np.arange(0, nb),
                                                       kptpair.kpt1,
                                                       kptpair.kpt2)

        # Check for nan's
        assert not np.isnan(n_nmG).any()
        assert not np.isnan(n_nmvG).any()
        if ol:
            assert not np.isnan(n2_nmv).any()

        # PAW correction test
        if ol:
            print('Checking PAW corrections')
            # Check that PAW-corrections are
            # are equal to nabla-PAW corrections
            G_Gv = pd.get_reciprocal_vectors()

            for id, atomdata in pair.calc.wfs.setups.setups.items():
                nabla_vii = atomdata.nabla_iiv.transpose((2, 0, 1))
                Q_vGii = two_phi_nabla_planewave_integrals(G_Gv, atomdata)
                ni = atomdata.ni
                Q_vGii.shape = (3, -1, ni, ni)
                msg = ('Planewave-nabla PAW corrections not equal ' +
                       'to nabla PAW corrections when q + G = 0!')
                assert nabla_vii == pytest.approx(Q_vGii[:, 0], abs=1e-10), msg

            # Check optical limit nabla matrix elements
            err = np.abs(n_nmvG[..., 0] - n2_nmv)
            maxerr = np.max(err)
            arg = np.unravel_index(np.argmax(err), err.shape)
            msg = 'G=0 pair densities wrong! ' + str(arg) + ' '
            assert maxerr == pytest.approx(0.0, abs=1e-10), msg

        # Check longitudinal part of matrix elements
        print('Checking continuity eq.')
        G_Gv = pd.get_reciprocal_vectors()
        G2_G = np.sum(G_Gv**2.0, axis=1)

        n2_nmG = np.diagonal(np.dot(G_Gv, n_nmvG), axis1=0, axis2=3).copy()
        if ol:
            n2_nmG[..., 0] = n_nmvG[..., 2, 0]

        # Left hand side and right hand side of
        # continuity eq.: d/dr x J + d/dt n = 0
        lhs_nmG = n2_nmG - G2_G[np.newaxis, np.newaxis] / 2. * n_nmG[...,
                                                                     2 * ol:]
        rhs_nmG = - deps_nm[..., np.newaxis] * n_nmG[..., 2 * ol:]

        err = np.abs(lhs_nmG - rhs_nmG)
        maxerr = np.max(err)
        arg = np.unravel_index(np.argmax(err), rhs_nmG.shape)
        msg = ('Calculated current does ' +
               'not fulfill the continuity equation! ' + str(arg) + ' ')
        assert maxerr == pytest.approx(0.0, abs=1e-3), msg



# Copyright (C) 2006 CSC-Scientific Computing Ltd.
# Please see the accompanying LICENSE file for further information.
import os
import sys
import re
from sysconfig import get_config_vars, get_platform
from glob import glob
from pathlib import Path
from stat import ST_MTIME


def mtime(path, name, mtimes):
    """Return modification time.

    The modification time of a source file is returned.  If one of its
    dependencies is newer, the mtime of that file is returned.
    This function fails if two include files with the same name
    are present in different directories."""

    include = re.compile(r'^#\s*include "(\S+)"', re.MULTILINE)

    if name in mtimes:
        return mtimes[name]
    t = os.stat(os.path.join(path, name))[ST_MTIME]
    for name2 in include.findall(open(os.path.join(path, name)).read()):
        path2, name22 = os.path.split(name2)
        if name22 != name:
            t = max(t, mtime(os.path.join(path, path2), name22, mtimes))
    mtimes[name] = t
    return t


def check_dependencies(sources):
    # Distutils does not do deep dependencies correctly.  We take care of
    # that here so that "python setup.py build_ext" always does the right
    # thing!
    mtimes = {}  # modification times

    # Remove object files if any dependencies have changed:
    plat = get_platform() + '-{maj}.{min}'.format(maj=sys.version_info[0],
                                                  min=sys.version_info[1])
    remove = False
    for source in sources:
        path, name = os.path.split(source)
        t = mtime(path + '/', name, mtimes)
        o = 'build/temp.%s/%s.o' % (plat, source[:-2])  # object file
        if os.path.exists(o) and t > os.stat(o)[ST_MTIME]:
            print('removing', o)
            os.remove(o)
            remove = True

    so = 'build/lib.{}/_gpaw.so'.format(plat)
    if os.path.exists(so) and remove:
        # Remove shared object C-extension:
        # print 'removing', so
        os.remove(so)


def write_configuration(define_macros, include_dirs, libraries, library_dirs,
                        extra_link_args, extra_compile_args,
                        runtime_library_dirs, extra_objects, mpicompiler,
                        mpi_libraries, mpi_library_dirs, mpi_include_dirs,
                        mpi_runtime_library_dirs, mpi_define_macros):

    # Write the compilation configuration into a file
    try:
        out = open('configuration.log', 'w')
    except IOError as x:
        print(x)
        return
    print("Current configuration", file=out)
    print("libraries", libraries, file=out)
    print("library_dirs", library_dirs, file=out)
    print("include_dirs", include_dirs, file=out)
    print("define_macros", define_macros, file=out)
    print("extra_link_args", extra_link_args, file=out)
    print("extra_compile_args", extra_compile_args, file=out)
    print("runtime_library_dirs", runtime_library_dirs, file=out)
    print("extra_objects", extra_objects, file=out)
    if mpicompiler is not None:
        print(file=out)
        print("Parallel configuration", file=out)
        print("mpicompiler", mpicompiler, file=out)
        print("mpi_libraries", mpi_libraries, file=out)
        print("mpi_library_dirs", mpi_library_dirs, file=out)
        print("mpi_include_dirs", mpi_include_dirs, file=out)
        print("mpi_define_macros", mpi_define_macros, file=out)
        print("mpi_runtime_library_dirs", mpi_runtime_library_dirs, file=out)
    out.close()


def build_interpreter(define_macros, include_dirs, libraries, library_dirs,
                      extra_link_args, extra_compile_args,
                      runtime_library_dirs, extra_objects,
                      mpicompiler, mpilinker, mpi_libraries, mpi_library_dirs,
                      mpi_include_dirs, mpi_runtime_library_dirs,
                      mpi_define_macros):

    # Build custom interpreter which is used for parallel calculations

    cfgDict = get_config_vars()
    plat = get_platform() + '-{maj}.{min}'.format(maj=sys.version_info[0],
                                                  min=sys.version_info[1])

    cfiles = glob('c/[a-zA-Z_]*.c') + ['c/bmgs/bmgs.c']
    cfiles += glob('c/xc/*.c')
    # Make build process deterministic (for "reproducible build" in debian)
    # XXX some of this is duplicated in setup.py!  Why do the same thing twice?
    cfiles.sort()

    sources = ['c/bc.c', 'c/mpi.c', 'c/_gpaw.c',
               'c/operators.c', 'c/woperators.c', 'c/transformers.c',
               'c/elpa.c',
               'c/blacs.c', 'c/utilities.c', 'c/xc/libvdwxc.c']
    objects = ' '.join(['build/temp.%s/' % plat + x[:-1] + 'o'
                        for x in cfiles])

    if not os.path.isdir('build/bin.{}/'.format(plat)):
        os.makedirs('build/bin.{}/'.format(plat))
    exefile = 'build/bin.{}/gpaw-python'.format(plat)

    libraries += mpi_libraries
    library_dirs += mpi_library_dirs
    define_macros += mpi_define_macros
    include_dirs += mpi_include_dirs
    runtime_library_dirs += mpi_runtime_library_dirs

    define_macros.append(('PARALLEL', '1'))
    define_macros.append(('GPAW_INTERPRETER', '1'))
    macros = ' '.join(['-D%s=%s' % x for x in define_macros if x[0].strip()])

    include_dirs.append(cfgDict['INCLUDEPY'])
    include_dirs.append(cfgDict['CONFINCLUDEPY'])
    includes = ' '.join(['-I' + incdir for incdir in include_dirs])

    library_dirs.append(cfgDict['LIBPL'])
    lib_dirs = ' '.join(['-L' + lib for lib in library_dirs])

    libs = ' '.join(['-l' + lib for lib in libraries if lib.strip()])
    # LIBDIR/INSTSONAME will point at the static library if that is how
    # Python was compiled:
    lib = Path(cfgDict['LIBDIR']) / cfgDict['INSTSONAME']
    if lib.is_file():
        libs += ' {}'.format(lib)
    else:
        libs += ' ' + cfgDict.get('BLDLIBRARY',
                                  '-lpython{}'.format(cfgDict['VERSION']))
    libs = ' '.join([libs, cfgDict['LIBS'], cfgDict['LIBM']])

    # Hack taken from distutils to determine option for runtime_libary_dirs
    if sys.platform[:6] == 'darwin':
        # MacOSX's linker doesn't understand the -R flag at all
        runtime_lib_option = '-L'
    elif sys.platform[:5] == 'hp-ux':
        runtime_lib_option = '+s -L'
    elif os.popen('mpicc --showme 2> /dev/null', 'r').read()[:3] == 'gcc':
        runtime_lib_option = '-Wl,-R'
    elif os.popen('mpicc -show 2> /dev/null', 'r').read()[:3] == 'gcc':
        runtime_lib_option = '-Wl,-R'
    else:
        runtime_lib_option = '-R'

    runtime_libs = ' '.join([runtime_lib_option + lib
                             for lib in runtime_library_dirs])

    extra_link_args.append(cfgDict['LDFLAGS'])

    if sys.platform in ['aix5', 'aix6']:
        extra_link_args.append(cfgDict['LINKFORSHARED'].replace(
            'Modules', cfgDict['LIBPL']))
    elif sys.platform == 'darwin':
        # On a Mac, it is important to preserve the original compile args.
        # This should probably always be done ?!?
        extra_compile_args.append(cfgDict['CFLAGS'])
        extra_link_args.append(cfgDict['LINKFORSHARED'])
    else:
        extra_link_args.append(cfgDict['LINKFORSHARED'])

    extra_compile_args.append('-fPIC')

    # Compile the parallel sources
    for src in sources:
        obj = 'build/temp.{}/'.format(plat) + src[:-1] + 'o'
        cmd = ('{} {} {} {} -o {} -c {} ').format(
               mpicompiler,
               macros,
               ' '.join(extra_compile_args),
               includes,
               obj,
               src)
        print(cmd)
        error = os.system(cmd)
        if error != 0:
            return error

    # Link the custom interpreter
    cmd = ('{} -o {} {} {} {} {} {} {}').format(
           mpilinker,
           exefile,
           objects,
           ' '.join(extra_objects),
           lib_dirs,
           libs,
           runtime_libs,
           ' '.join(extra_link_args))

    print(cmd)
    error = os.system(cmd)
    return error



from ase.parallel import paropen
from ase.units import Hartree
from gpaw.xc.rpa import RPACorrelation

rpa = RPACorrelation('N.gpw', nblocks=8, truncation='wigner-seitz',
                     txt='rpa_N.txt')
E1_i = rpa.calculate(ecut=400)

rpa = RPACorrelation('N2.gpw', nblocks=8, truncation='wigner-seitz',
                     txt='rpa_N2.txt')
E2_i = rpa.calculate(ecut=400)
ecut_i = rpa.ecut_i

f = paropen('rpa_N2.dat', 'w')
for ecut, E1, E2 in zip(ecut_i, E1_i, E2_i):
    print(ecut * Hartree, E2 - 2 * E1, file=f)
f.close()



from myqueue.workflow import run


def workflow():
    runs = [run(script='ruslab.py', cores=8, tmax='10h'),
            run(script='ruslab.py', args=['H'], cores=8, tmax='10h'),
            run(script='ruslab.py', args=['N'], cores=8, tmax='10h'),
            run(script='ruslab.py', args=['O'], cores=16, tmax='15h'),
            run(script='molecules.py', cores=8, tmax='20m')]
    run(script='results.py', deps=runs)



parallel_python_interpreter = True
compiler = 'cc'
mpicompiler = 'cc'
mpilinker = 'cc'
scalapack = True
libxc = '/usr/common/software/libxc/4.2.3/gnu/haswell'
include_dirs += [libxc + '/include']
library_dirs += [libxc + '/lib']
extra_link_args += [f'-Wl,-rpath={libxc}/lib']
extra_compile_args += ['-O2']
if 'xc' not in libraries:
    libraries.append('xc')
# these are in the cray wrapper
if 'blas' in libraries:
    libraries.remove('blas')
if 'lapack' in libraries:
    libraries.remove('lapack')

if scalapack:
    define_macros += [('GPAW_NO_UNDERSCORE_CBLACS', '1')]
    define_macros += [('GPAW_NO_UNDERSCORE_CSCALAPACK', '1')]



from ase import Atoms
from ase.parallel import paropen
from gpaw import GPAW
from gpaw import PW

resultfile = paropen('H.ralda.DFT_corr_energies.txt', 'w')
resultfile.write('DFT Correlation energies for H atom\n')

H = Atoms('H', [(0, 0, 0)])
H.set_pbc(True)
H.center(vacuum=2.0)
calc = GPAW(mode=PW(300, force_complex_dtype=True),
            parallel={'domain': 1},
            hund=True,
            txt='H.ralda_01_lda.output.txt',
            xc='LDA')

H.calc = calc
E_lda = H.get_potential_energy()
E_c_lda = -calc.get_xc_difference('LDA_X')

resultfile.write(f'LDA correlation: {E_c_lda} eV')
resultfile.write('\n')

calc.diagonalize_full_hamiltonian()
calc.write('H.ralda.lda_wfcs.gpw', mode='all')



# creates: 2sigma.png, co_wavefunctions.png
import numpy as np
import matplotlib.pyplot as plt
from ase import Atoms
from ase.units import Bohr

from gpaw import GPAW
from gpaw.spherical_harmonics import Y

L = 6.0
d = 1.13
co = Atoms('CO',
           cell=[L, L, L],
           positions=[(0, 0, 0), (d, 0, 0)])
co.center()
co.calc = GPAW(mode='lcao',
               txt='CO.txt')
e = co.get_potential_energy()
print(co.positions[:, 0] - L / 2)

dpi = 100
C = 'g'
N = 100
for a, pp in enumerate(co.calc.wfs.setups):
    rc = max(pp.rcut_j)
    print(pp.rcut_j)
    x = np.linspace(-rc, rc, 2 * N + 1)
    P_i = co.calc.wfs.kpt_qs[0][0].projections[a][1] / Bohr**1.5
    phi_i = np.empty((len(P_i), len(x)))
    phit_i = np.empty((len(P_i), len(x)))
    i = 0
    for l, phi_g, phit_g in zip(pp.l_j, pp.data.phi_jg, pp.data.phit_jg):
        f = pp.rgd.spline(phi_g, rc + 0.3, l).map(x[N:]) * x[N:]**l
        ft = pp.rgd.spline(phit_g, rc + 0.3, l).map(x[N:]) * x[N:]**l
        for m in range(2 * l + 1):
            ll = l**2 + m
            phi_i[i, N:] = f * Y(ll, 1, 0, 0)
            phi_i[i, N::-1] = f * Y(ll, -1, 0, 0)
            phit_i[i, N:] = ft * Y(ll, 1, 0, 0)
            phit_i[i, N::-1] = ft * Y(ll, -1, 0, 0)
            i += 1
    x0 = co.positions[a, 0] - L / 2
    symbol = co.symbols[a]
    print(symbol, x0, rc)
    plt.plot([x0], [0], 'o', ms=dpi * rc * 2 / 2.33 * 1.3 * Bohr,
             mfc='None', label='_nolegend_')
    plt.plot(x * Bohr + x0, P_i.dot(phit_i), C + '-', lw=1,
             label=r'$\tilde{\psi}^%s$' % symbol)
    plt.plot(x * Bohr + x0, P_i.dot(phi_i), C + '-', lw=2,
             label=r'$\psi^%s$' % symbol)
    C = 'r'

psit = co.calc.get_pseudo_wave_function(band=1)
n = len(psit)
psit2 = psit[:, :, n // 2]
psit1 = psit2[:, n // 2]

x = np.linspace(-L / 2, L / 2, n, endpoint=False)
plt.plot(x, psit1, 'bx', mew=2, label=r'$\tilde{\psi}$')

plt.legend(loc='best')
plt.xlabel('x [Å]')
plt.ylabel(r'$\psi$')
plt.ylim(ymin=-2, ymax=2)
# plt.show()
plt.savefig('co_wavefunctions.png', dpi=dpi)

fig = plt.figure()
ax = fig.add_subplot(111)
m = abs(psit2).max() * 1.1
cax = ax.contour(x, x, psit2.T, np.linspace(-m, m, 31))
ax.text(-d / 2, 0, 'C', ha='center', va='center')
ax.text(d / 2, 0, 'O', ha='center', va='center')
cbar = fig.colorbar(cax)
ax.set_xlabel('x (Angstrom)')
ax.set_ylabel('y (Angstrom)')
# plt.show()
fig.savefig('2sigma.png')



from sys import argv
from ase.build import hcp0001, add_adsorbate
from ase.constraints import FixAtoms
from ase.optimize.lbfgs import LBFGS
from gpaw import GPAW, Mixer, FermiDirac

tag = 'Ru001'

adsorbate_heights = {'H': 1.0, 'N': 1.108, 'O': 1.257}

slab = hcp0001('Ru', size=(2, 2, 4), a=2.72, c=1.58 * 2.72, vacuum=7.0,
               orthogonal=True)
slab.center(axis=2)

if len(argv) > 1:
    adsorbate = argv[1]
    tag = adsorbate + tag
    add_adsorbate(slab, adsorbate, adsorbate_heights[adsorbate], 'hcp')

slab.set_constraint(FixAtoms(mask=slab.get_tags() >= 3))

calc = GPAW(xc='PBE',
            h=0.2,
            mixer=Mixer(0.1, 5, weight=100.0),
            occupations=FermiDirac(width=0.1),
            kpts=[4, 4, 1],
            eigensolver='cg',
            txt=tag + '.txt')
slab.calc = calc
  
opt = LBFGS(slab, logfile=tag + '.log', trajectory=tag + '.traj')
opt.run(fmax=0.05)
calc.write(tag + '.gpw')



import sys
from ase import Atoms
from ase.parallel import paropen
from ase.data.s22 import data
from ase.calculators.vdwcorrection import vdWTkatchenko09prl
from gpaw import GPAW, FermiDirac
from gpaw.cluster import Cluster
from gpaw.analyse.hirshfeld import HirshfeldPartitioning
from gpaw.analyse.vdwradii import vdWradii
try:
    from dftd4 import D4_model
except ModuleNotFoundError:
    pass

h = 0.18
box = 4.

xc = 'TS09'
if len(sys.argv) > 1:
    xc = sys.argv[1]

f = paropen('energies_' + xc + '.dat', 'w')
print('# h=', h, file=f)
print('# box=', box, file=f)
print('# molecule E[1]  E[2]  E[1+2]  E[1]+E[2]-E[1+2]', file=f)
for molecule in data:
    print(molecule, end=' ', file=f)
    ss = Cluster(Atoms(data[molecule]['symbols'],
                       data[molecule]['positions']))
    # split the structures
    s1 = ss.find_connected(0)
    s2 = ss.find_connected(-1)
    assert len(ss) == len(s1) + len(s2)
    if xc == 'TS09' or xc == 'TPSS' or xc == 'M06-L' or xc == 'dftd4':
        c = GPAW(xc='PBE', h=h, nbands=-6, occupations=FermiDirac(width=0.1))
    else:
        c = GPAW(xc=xc, h=h, nbands=-6, occupations=FermiDirac(width=0.1))
    E = []
    for s in [s1, s2, ss]:
        s.calc = c
        s.minimal_box(box, h=h)
        if xc == 'TS09':
            s.get_potential_energy()
            cc = vdWTkatchenko09prl(HirshfeldPartitioning(c),
                                    vdWradii(s.get_chemical_symbols(), 'PBE'))
            s.calc = cc
        elif xc == 'dftd4':
            s.get_potential_energy()
            cc = D4_model(xc='PBE', calc=c)
            s.calc = cc
        if xc == 'TPSS' or xc == 'M06-L':
            ene = s.get_potential_energy()
            ene += c.get_xc_difference(xc)
            E.append(ene)
        else:
            E.append(s.get_potential_energy())
    print(E[0], E[1], E[2], end=' ', file=f)
    print(E[0] + E[1] - E[2], file=f)
    f.flush()
f.close()



"""NiO in an antiferromagnetic configuration.

See:
    Quasiparticle energy bands of NiO in the GW approximation
    Je-Luen Li, G.-M. Rignanese, and Steven G. Louie
    10.1103/PhysRevB.71.193102
"""
from ase import Atoms
from gpaw import GPAW

a = 4.1767  # lattice constants
b = a / 2
m = 2.0

atoms = Atoms('Ni2O2',
              pbc=True,
              cell=[(b, b, 0),
                    (0, b, b),
                    (a, 0, a)],
              positions=[(0, 0, 0),
                         (b, 0, b),
                         (b, 0, 0),
                         (a, 0, b)],
              magmoms=[m, -m, 0, 0])
atoms.write('nio.json')

atoms.calc = GPAW(mode={'name': 'pw', 'ecut': 600},
                  txt='nio.txt',
                  kpts={'density': 3.0, 'gamma': True},
                  xc='PBE')

e = atoms.get_potential_energy()

calc = atoms.calc.fixed_density(nbands=100)
calc.write('nio.gpw', mode='all')



from ase.transport.calculators import TransportCalculator
import numpy as np
import pickle
import pylab

# Principal layer size
# Uncomment this line if going back to gpawtransport again
# pl = 4 * 9 # 9 is the number of bf per Pt atom (basis=szp), see below

# Read in the hamiltonians
h, s = pickle.load(open('scat_hs.pickle', 'rb'))
# Uncomment this line if going back to gpawtransport again
# h, s = h[pl:-pl, pl:-pl], s[pl:-pl, pl:-pl]
h1, s1 = pickle.load(open('lead1_hs.pickle', 'rb'))
h2, s2 = pickle.load(open('lead2_hs.pickle', 'rb'))

tcalc = TransportCalculator(h=h, h1=h1, h2=h2,  # hamiltonian matrices
                            s=s, s1=s1, s2=s2,  # overlap matrices
                            align_bf=1)        # align the Fermi levels

# Calculate the conductance (the energy zero corresponds to the Fermi level)
tcalc.set(energies=[0.0])
G = tcalc.get_transmission()[0]
print(f'Conductance: {G:.2f} 2e^2/h')

# Determine the basis functions of the two Hydrogen atoms and subdiagonalize
Pt_N = 5    # Number of Pt atoms on each side in the scattering region
Pt_nbf = 15  # number of bf per Pt atom (basis=szp)
H_nbf = 4   # number of bf per H atom (basis=szp)
bf_H1 = Pt_nbf * Pt_N
bfs = range(bf_H1, bf_H1 + 2 * H_nbf)
h_rot, s_rot, eps_n, vec_jn = tcalc.subdiagonalize_bfs(bfs)
for n in range(len(eps_n)):
    print("bf %i corresponds to the eigenvalue %.2f eV" % (bfs[n], eps_n[n]))

# Switch to the rotated basis set
tcalc.set(h=h_rot, s=s_rot)

# plot the transmission function
tcalc.set(energies=np.arange(-8, 4, 0.1))
T = tcalc.get_transmission()
pylab.plot(tcalc.energies, T)
pylab.title('Transmission function')
pylab.show()

# ... and the projected density of states (pdos) of the H2 molecular orbitals
tcalc.set(pdos=bfs)
pdos_ne = tcalc.get_pdos()
pylab.plot(tcalc.energies, pdos_ne[0], label='bonding')
pylab.plot(tcalc.energies, pdos_ne[1], label='anti-bonding')
pylab.title('Projected density of states')
pylab.legend()
pylab.show()

# Cut the coupling to the anti-bonding orbital.
print('Cutting the coupling to the renormalized molecular state at %.2f eV' % (
    eps_n[1]))
h_rot_cut, s_rot_cut = tcalc.cutcoupling_bfs([bfs[1]])
tcalc.set(h=h_rot_cut, s=s_rot_cut)
pylab.plot(tcalc.energies, tcalc.get_transmission())
pylab.title('Transmission without anti-bonding orbital')
pylab.show()

# Cut the coupling to the bonding-orbital.
print('Cutting the coupling to the renormalized molecular state at %.2f eV' % (
    eps_n[0]))
tcalc.set(h=h_rot, s=s_rot)
h_rot_cut, s_rot_cut = tcalc.cutcoupling_bfs([bfs[0]])
tcalc.set(h=h_rot_cut, s=s_rot_cut)
pylab.plot(tcalc.energies, tcalc.get_transmission())
pylab.title('Transmission without bonding orbital')
pylab.show()



if 1:  # example: set eigensolver/mixer
    from gpaw.eigensolvers.davidson import Davidson
    from gpaw import MixerSum

    def calc(atoms):
        atoms.calc.set(eigensolver=Davidson(), mixer=MixerSum(0.05, 5, 50))



from gpaw.densities import add
from gpaw.spline import Spline
from gpaw.core import UniformGrid
import numpy as np


def test_ae_density():
    n = 10
    n_sR = UniformGrid(cell=[1, 1, 1], size=[n, n, n], pbc=True).zeros(1)
    D_sii = np.zeros((1, 4, 4))
    D_sii[0, 0, 0] = 1.0
    D_sii[0, 1, 1] = 1.0
    rc = 0.5
    phi0 = Spline(0, rc, [1, 0])
    phit0 = Spline(0, rc, [2, 0])
    phi1 = Spline(1, rc, [1, 0])
    phit1 = Spline(1, rc, [2, 0])
    add([0.5, 0.5, 0.5], n_sR, [phi0, phi1], [phit0, phit1], rc, D_sii)
    y, v = n_sR.xy(0, 5, ..., 5)
    v *= 4 * np.pi
    y = abs(y - 0.5)
    v0 = -9 * phi1.map(y)**2 * y**2 - 3 * phi0.map(y)**2
    assert abs(v - v0).max() < 1e-14



from pathlib import Path
from ase.utils import IOContext
from ase.utils.timing import Timer

from gpaw import disable_dry_run
from gpaw import GPAW
import gpaw.mpi as mpi


def new_context(txt, world, timer):
    timer = timer or Timer()
    return ResponseContext(txt=txt, timer=timer, world=world)


def calc_and_context(calc, txt, world, timer):
    context = new_context(txt, world, timer)
    with context.timer('Read ground state'):
        try:
            path = Path(calc)
        except TypeError:
            pass
        else:
            print('Reading ground state calculation:\n  %s' % path,
                  file=context.fd)
            with disable_dry_run():
                calc = GPAW(path, communicator=mpi.serial_comm)

    assert calc.wfs.world.size == 1
    return calc, context


class ResponseContext:
    def __init__(self, txt, timer, world):
        self.iocontext = IOContext()
        self.fd = self.iocontext.openfile(txt, world)
        self.timer = timer
        self.world = world

    def close(self):
        self.iocontext.close()

    def __del__(self):
        self.close()

    def with_txt(self, txt):
        return new_context(txt=txt, world=self.world, timer=self.timer)



import os
import shutil
from pathlib import Path
from myqueue.workflow import run


def workflow():
    if os.getenv('AGTS_FILES'):
        dir = Path(os.getenv('AGTS_FILES'))
        file = Path('organometal.db')
        if not file.is_file():
            shutil.copyfile(dir / file, file)
    run(script='machinelearning.py', tmax='8h')



# web-page: gaps.csv
from ase.io import read


with open('gaps.csv', 'w') as fd:
    gaps = []
    for name in ['no_u', 'normalized_u', 'not_normalized_u']:
        n = read(name + '.txt')
        gap = n.calc.get_eigenvalues(spin=1)[1] - \
            n.calc.get_eigenvalues(spin=0)[1]
        gaps.append(gap)
        print(f"{name.replace('_', ' ').replace('u', 'U')}, {gap:.3f}",
              file=fd)

assert abs(gaps[1] - gaps[0] - 6.0) < 0.8



from gpaw.utilities import h2gpts
from gpaw import GPAW
from ase.build import fcc100
ncpus = 4
atoms = fcc100(symbol='Ni', size=(1, 1, 9), a=3.52, vacuum=5.5)
atoms.set_initial_magnetic_moments([0.6] * len(atoms))
gpts = h2gpts(0.18, atoms.cell, idiv=8)
atoms.calc = GPAW(gpts=gpts,
                  kpts=(8, 8, 1),
                  xc='PBE')



from ase import Atoms
from gpaw import GPAW, PW, FermiDirac

# Setup up bulk NiO in an antiferromagnetic configuration:
a = 4.19  # lattice constants
b = a / 2**0.5
m = 2.0
atoms = Atoms('Ni2O2',
              pbc=True,
              cell=(b, b, a),
              positions=[(0, 0, 0),
                         (b / 2, b / 2, a / 2),
                         (0, 0, a / 2),
                         (b / 2, b / 2, 0)],
              magmoms=[m, -m, 0, 0])

k = 2  # number of k-points
atoms.calc = GPAW(mode=PW(400),
                  occupations=FermiDirac(width=0.05),
                  setups={'Ni': ':d,6.0'},  # U=6 eV for Ni d orbitals
                  txt='nio.txt',
                  kpts=(k, k, k),
                  xc='PBE')
e = atoms.get_potential_energy()



from ase import io
from ase.io.bader import attach_charges

hirsh = io.read('Hirshfeld.traj')

bader = hirsh.copy()
attach_charges(bader)

print('atom Hirshfeld Bader')
for ah, ab in zip(hirsh, bader):
    assert ah.symbol == ab.symbol
    print(f'{ah.symbol:4s} {ah.charge:9.2f} {ab.charge:5.2f}')



from myqueue.workflow import run


def workflow():
    with run(script='gs_MoS2.py', cores=16, tmax='25m'):
        r1 = run(script='bb_MoS2.py', cores=16, tmax='20h')
    with run(script='gs_WSe2.py', cores=16, tmax='25m'):
        r2 = run(script='bb_WSe2.py', cores=16, tmax='20h')
    with r1, r2:
        with run(script='interpolate_bb.py'):
            run(script='interlayer.py')



from myqueue.workflow import run


def workflow():
    run(script='benzene-dimer-T-shaped.py', cores=48, tmax='20h')
    run(script='adenine-thymine_complex_stack.py', cores=4, tmax='2h')
    run(script='graphene_hirshfeld.py')



import numpy as np
from gpaw import GPAW
from gpaw.spinorbit import get_parity_eigenvalues

ik = 0

calc = GPAW('high_sym.gpw', txt=None)
Nv = int(calc.get_number_of_electrons() / 2)
r_v = [0, 0, 0]
multiplicities = np.array([1, 3, 3, 1])

print()
print('Calculating parities without spin-orbit coupling')
P_i = []
for ik, X in zip(range(4), ['G', 'L', 'F', 'Z']):
    ps = get_parity_eigenvalues(calc, ik=ik, spin_orbit=False, bands=range(Nv),
                                inversion_center=r_v)
    P = np.prod(ps[:Nv])
    P_i.append(P)
    print(f'    Parity at {X}:', P)
print()
print('Calculating parities with spin-orbit coupling')
Pso_i = []
for ik, X in zip(range(4), ['G', 'L', 'F', 'Z']):
    ps = get_parity_eigenvalues(calc, ik=ik, spin_orbit=True,
                                bands=range(Nv + 10),
                                inversion_center=r_v, deg_tol=1.0e-8)
    P = np.prod(ps[:Nv])
    Pso_i.append(P)
    print(f'    Parity at {X}:', P)
print()
print('Z2 index no spinorbit:',
      (1 - np.prod(np.array(P_i)**multiplicities)) / 2)
print('Z2 index spinorbit',
      (1 - np.prod(np.array(Pso_i)**multiplicities)) / 2)



"""A proton in an electric field."""
import pytest
import numpy as np
from ase import Atoms
from ase.units import Hartree, Bohr

from gpaw import GPAW
from gpaw.external import ConstantElectricField
from gpaw.external import static_polarizability


def test_ext_potential_constant_e_field(in_tmp_dir):
    h = Atoms('H')
    h.center(vacuum=2.5)
    h.calc = GPAW(external=ConstantElectricField(1.0),  # 1 V / Ang
                  charge=1,
                  txt='h.txt')
    e = h.get_potential_energy()
    f1 = h.get_forces()[0, 2]
    h[0].z += 0.001
    de = h.get_potential_energy() - e
    f2 = -de / 0.001
    print(f1, f2)
    assert abs(f1 - 1) < 1e-4
    assert abs(f2 - 1) < 5e-3

    # Check writing and reading:
    h.calc.write('h')
    vext = GPAW('h', txt=None).hamiltonian.vext
    assert abs(vext.field_v[2] - 1.0 * Bohr / Hartree) < 1e-13


def test_polarizability(in_tmp_dir):
    H2 = Atoms('H2', positions=[(0, 0, 0), (0.7, 0, 0)])
    H2.center(vacuum=2.5)
    H2.calc = GPAW()

    strength = 0.1  # V/Ang
    alpha_cc = static_polarizability(H2, strength)

    # make sure no external potential is left over
    assert H2.calc.parameters.external is None

    assert alpha_cc.shape == (3, 3)
    assert alpha_cc == pytest.approx(
        np.diag([6.48529231e-02, 4.61303856e-2, 4.61303856e-2]))



# creates: lines.png
import datetime as dt
import os
import subprocess

import numpy as np
import pylab as pl


def count(dir, pattern):
    if not os.path.isdir(dir):
        return 0
    files = subprocess.check_output(f'find {dir} -name {pattern}',
                                    shell=True).decode().split()[:-1]
    if not files:
        return 0
    out = subprocess.check_output(f"wc -l {' '.join(files)} | tail -1",
                                  shell=True)
    return int(out.split()[0])


def polygon(x, y1, y2, *args, **kwargs):
    x = pl.concatenate((x, x[::-1]))
    y = pl.concatenate((y1, y2[::-1]))
    pl.fill(x, y, *args, **kwargs)


def plot_count(dpi=70):
    year, month, f, c, py, test, doc, rst = np.loadtxt('lines.data').T
    date = year + (month - 1) / 12

    fig = pl.figure(1, figsize=(10, 5), dpi=dpi)
    fig.add_subplot(111)
    polygon(date, c + py + test + doc, c + py + test + doc + rst,
            facecolor='m', label='Documentation (.rst)')
    polygon(date, c + py + test, c + py + test + doc,
            facecolor='c', label='Documentation (.py)')
    polygon(date, c + py, c + py + test,
            facecolor='y', label='Tests (.py)')
    polygon(date, c, c + py,
            facecolor='g', label='Python-code (.py) ')
    polygon(date, f, c,
            facecolor='r', label='C-code (.c, .h)')
    polygon(date, f, f,
            facecolor='b', label='Fortran-code')

    pl.axis('tight')
    pl.legend(loc='upper left')
    pl.title('Number of lines')
    pl.savefig('lines.png', dpi=dpi)


def count_lines():
    now = dt.date.today()
    stop = now.year, now.month
    year = 2005
    month = 11

    with open('lines.data', 'w') as fd:
        while (year, month) <= stop:
            hash = subprocess.check_output(
                'git rev-list -n 1 --before="{}-{}-01 12:00" master'
                .format(year, month), shell=True).strip()
            print(year, month, hash)
            subprocess.call(['git', 'checkout', hash])
    
            c = count('c', r'\*.[ch]')
            py = count('.', r'\*.py')
            test = count('gpaw/test', r'\*.py')
            test += count('test', r'\*.py')
            doc = count('doc', r'\*.py')
            py -= test + doc  # avoid double counting
            rst = count('.', r'\*.rst')
            print(year, month, 0, c, py, test, doc, rst, file=fd)
            month += 1
            if month == 13:
                month = 1
                year += 1


plot_count()



from myqueue.workflow import run


def workflow():
    run(script='plot_freq.py')
    with run(script='silicon_ABS_simpleversion.py'):
        run(script='plot_silicon_ABS_simple.py')

    with run(script='silicon_ABS.py', cores=16, tmax='1h'):
        run(script='plot_ABS.py')

    with run(script='aluminum_EELS.py', cores=8, tmax='1h'):
        run(script='plot_aluminum_EELS_simple.py')

    with run(script='graphite_EELS.py', cores=8, tmax='1h'):
        run(script='plot_EELS.py')

    run(script='tas2_dielectric_function.py', cores=8, tmax='15m')
    run(script='graphene_dielectric_function.py', cores=8, tmax='15m')



from typing import Any, Union, Sequence
import numpy as np

if [int(x) for x in np.__version__.split('.')] >= [1, 20]:
    from numpy.typing import ArrayLike, DTypeLike
else:
    ArrayLike = Any  # type: ignore
    DTypeLike = Any  # type: ignore

ArrayLike1D = ArrayLike
ArrayLike2D = ArrayLike

ArrayND = np.ndarray
Array1D = ArrayND
Array2D = ArrayND
Array3D = ArrayND
Array4D = ArrayND

# Used for sequences of three numbers:
Vector = Union[Sequence[float], Array1D]
IntVector = Union[Sequence[int], Array1D]



# Convert static library specs from EasyBuild to GPAW
def static_eblibs_to_gpawlibs(lib_specs):
    return [libfile[3:-2] for libfile in os.getenv(lib_specs).split(',')]

# Clean out any autodetected things, we only want the EasyBuild
# definitions to be used.
libraries = []
include_dirs = []

# Use EasyBuild fftw from the active toolchain
fftw = os.getenv('FFT_STATIC_LIBS')
if fftw:
    libraries += static_eblibs_to_gpawlibs('FFT_STATIC_LIBS')

# Use ScaLAPACK from the active toolchain
scalapack = os.getenv('SCALAPACK_STATIC_LIBS')
if scalapack:
    libraries += static_eblibs_to_gpawlibs('SCALAPACK_STATIC_LIBS')

# Add EasyBuild LAPACK/BLAS libs
libraries += static_eblibs_to_gpawlibs('LAPACK_STATIC_LIBS')
libraries += static_eblibs_to_gpawlibs('BLAS_STATIC_LIBS')

# LibXC:
# Use EasyBuild libxc
libxc = os.getenv('EBROOTLIBXC')
if libxc:
    include_dirs.append(os.path.join(libxc, 'include'))
    libraries.append('xc')

# libvdwxc:
# Use EasyBuild libvdwxc
libvdwxc = os.getenv('EBROOTLIBVDWXC')
if libvdwxc:
    include_dirs.append(os.path.join(libvdwxc, 'include'))
    libraries.append('vdwxc')

# ELPA:
# Use EasyBuild ELPA if loaded
elpa = os.getenv('EBROOTELPA')
if elpa:
    libraries += ['elpa']
    elpaversion = os.path.basename(elpa).split('-')[0]
    library_dirs.append(os.path.join(elpa, 'lib'))
    extra_link_args += [f'-Wl,-rpath={elpa}/lib']
    include_dirs.append(os.path.join(elpa, 'include', 'elpa-'+elpaversion))

# Now add a EasyBuild "cover-all-bases" library_dirs
library_dirs = os.getenv('LD_LIBRARY_PATH').split(':')



import runpy
globals().update(runpy.run_path('workflow.py'))


@asr.parametrize_glob('*/material')
def workflow(material):
    calculator = {'mode': 'pw', 'kpts': {'density': 1.0}, 'txt': 'gpaw.txt'}
    return MyWorkflow(atoms=material, calculator=calculator)



def workflow():
    from myqueue.workflow import run
    with run(script='si.pbe.py'):
        run(script='si_pbe_exx.py', cores=4, tmax='15m')
    with run(script='si.rpa_init_pbe.py'):
        run(script='si.rpa.py', cores=4, tmax='15m')



from ase.io import read
e0 = read('bulk.txt').get_potential_energy()
e = read('surface.txt').get_potential_energy()
a = 1.6
L = 10 * a
sigma = (e - L / a * e0) / 2 / a**2
print(f'{1000 * sigma:.2f} mev/Ang^2')
print(f'{sigma / 6.24150974e-05:.1f} erg/cm^2')
assert abs(sigma - 0.0054) < 0.0001



from ase.build import fcc111, molecule
from ase.units import Pascal, m

from gpaw.solvation.sjm import SJM, SJMPower12Potential
from gpaw.solvation import (
    EffectivePotentialCavity,
    LinearDielectric,
    GradientSurface,
    SurfaceInteraction)

# Build a tiny gold slab with a single water molecule above.
atoms = fcc111('Au', size=(1, 1, 3))
atoms.center(axis=2, vacuum=12.)
atoms.translate([0., 0., -4.])
water = molecule('H2O')
water.rotate('y', 90.)
water.positions += atoms[2].position + (0., 0., 4.4) - water[0].position
atoms.extend(water)

# Solvated jellium parameters.
sj = {'target_potential': 4.2}  # Desired potential

# Implicit solvent parameters (to SolvationGPAW).
epsinf = 78.36  # dielectric constant of water at 298 K
gamma = 18.4 * 1e-3 * Pascal * m
cavity = EffectivePotentialCavity(
    effective_potential=SJMPower12Potential(H2O_layer=True),
    temperature=298.15,  # K
    surface_calculator=GradientSurface())
dielectric = LinearDielectric(epsinf=epsinf)
interactions = [SurfaceInteraction(surface_tension=gamma)]

# The calculator
calc = SJM(
    # General GPAW parameters.
    txt='Au111.txt',
    gpts=(16, 16, 136),
    kpts=(9, 9, 1),
    xc='PBE',
    maxiter=1000,
    # Solvated jellium parameters.
    sj=sj,
    # Implicit solvent parameters.
    cavity=cavity,
    dielectric=dielectric,
    interactions=interactions)
atoms.calc = calc

# Run the calculation.
atoms.get_potential_energy()
atoms.write('Au111.traj')
calc.write_sjm_traces(path='sjm_traces.out')  # .out for .gitignore




from ase import Atom, Atoms

from gpaw import GPAW
from gpaw.test import equal


def test_aedensity(in_tmp_dir):
    try:
        calc = GPAW('NaCl.gpw')
        NaCl = calc.get_atoms()
        e = NaCl.get_potential_energy()
    except IOError:
        h = 0.21  # gridspacing
        a = [6.5, 6.5, 7.7]  # unit cell
        d = 2.3608  # experimental bond length

        NaCl = Atoms([Atom('Na', [0, 0, 0]),
                      Atom('Cl', [0, 0, d])],
                     pbc=False, cell=a)
        NaCl.center()
        calc = GPAW(h=h, xc='LDA', nbands=5,
                    setups={'Na': '1'},
                    convergence={'eigenstates': 1e-6}, spinpol=1)

        NaCl.calc = calc
        e = NaCl.get_potential_energy()
        calc.write('NaCl.gpw')

    equal(e, -4.907, 0.002)

    dv = NaCl.get_volume() / calc.get_number_of_grid_points().prod()
    nt1 = calc.get_pseudo_density(gridrefinement=1)
    Zt1 = nt1.sum() * dv
    nt2 = calc.get_pseudo_density(gridrefinement=2)
    Zt2 = nt2.sum() * dv / 8
    print('Integral of pseudo density:', Zt1, Zt2)
    equal(Zt1, Zt2, 1e-12)

    for gridrefinement in [1, 2, 4]:
        n = calc.get_all_electron_density(gridrefinement=gridrefinement)
        Z = n.sum() * dv / gridrefinement**3
        print('Integral of all-electron density:', Z)
        equal(Z, 28, 1e-5)



from myqueue.workflow import run
from pathlib import Path


def workflow():
    with run(script='H2O_ir.py'):
        run(script='H2O_ir_summary.py')
        with run(script='H2O_rraman_calc.py', cores=4, tmax='1h'):
            run(script='H2O_rraman_spectrum.py')
            with run(script='H2O_rraman_summary.py'):
                run(function=check)


def check():
    """Read result and make sure it's OK."""
    lines = Path('H2O_rraman_summary.txt').read_text().splitlines()
    for line in lines:
        if line.strip().startswith('8'):
            _, frequency, _, intensity = (float(x) for x in line.split())
            assert abs(frequency - 474) < 1
            assert abs(intensity - 57) < 1
            return
    raise ValueError



from myqueue.workflow import run


def workflow():
    with run(script='gs.py', cores=48, tmax='30m'):
        with run(script='td.py', cores=48, tmax='4h'):
            with run(script='spec.py') as spec:
                run(script='plot_spec.py')

    with run(script='basis.py', folder='mybasis'):
        with run(script='gs.py', cores=48, tmax='30m', folder='mybasis'):
            with run(script='td.py', cores=48, tmax='4h', folder='mybasis'):
                spec_my = run(script='spec.py', folder='mybasis')

    with run(script='gs.py', cores=48, tmax='2h', folder='dzp'):
        with run(script='td.py', cores=40, tmax='5h', folder='dzp'):
            spec_dzp = run(script='spec.py', folder='dzp')

    run(script='plot_spec_basis.py', deps=[spec, spec_my, spec_dzp])



from __future__ import annotations

import warnings
from pathlib import Path
from types import SimpleNamespace
from typing import IO, Any, Union

from ase import Atoms
from ase.units import Bohr, Ha

from gpaw import __version__
from gpaw.new import Timer, cached_property
from gpaw.new.builder import builder as create_builder
from gpaw.new.calculation import DFTCalculation, DFTState, units
from gpaw.new.gpw import read_gpw, write_gpw
from gpaw.new.input_parameters import InputParameters
from gpaw.new.logger import Logger
from gpaw.new.pw.fulldiag import diagonalize
from gpaw.new.xc import XC
from gpaw.typing import Array1D, Array2D, Array3D
from gpaw.utilities import pack
from gpaw.utilities.memory import maxrss


def GPAW(filename: Union[str, Path, IO[str]] = None,
         **kwargs) -> ASECalculator:
    """Create ASE-compatible GPAW calculator."""
    params = InputParameters(kwargs)
    txt = params.txt
    if txt == '?':
        txt = '-' if filename is None else None
    world = params.parallel['world']
    log = Logger(txt, world)

    if filename is not None:
        assert set(kwargs) <= {'txt', 'parallel', 'communicator'}, kwargs
        atoms, calculation, params, _ = read_gpw(filename, log,
                                                 params.parallel)
        return ASECalculator(params, log, calculation, atoms)

    write_header(log, world, params)
    return ASECalculator(params, log)


class ASECalculator:
    """This is the ASE-calculator frontend for doing a GPAW calculation."""
    def __init__(self,
                 params: InputParameters,
                 log: Logger,
                 calculation=None,
                 atoms=None):
        self.params = params
        self.log = log
        self.calculation = calculation

        self.atoms = atoms
        self.timer = Timer()

    def __repr__(self):
        params = []
        for key, value in self.params.items():
            val = repr(value)
            if len(val) > 40:
                val = '...'
            params.append((key, val))
        p = ', '.join(f'{key}: {val}' for key, val in params)
        return f'ASECalculator({p})'

    def calculate_property(self, atoms: Atoms, prop: str) -> Any:
        """Calculate (if not already calculated) a property.

        Must be one of

        * energy
        * forces
        * stress
        * magmom
        * magmoms
        * dipole
        """
        if self.calculation is not None:
            changes = compare_atoms(self.atoms, atoms)
            if changes & {'numbers', 'pbc', 'cell'}:
                # Start from scratch:
                if 'numbers' not in changes:
                    # Remember magmoms if there are any:
                    magmom_a = self.calculation.results.get('magmoms')
                    if magmom_a is not None:
                        atoms = atoms.copy()
                        atoms.set_initial_magnetic_moments(magmom_a)
                self.calculation = None

        if self.calculation is None:
            self.create_new_calculation(atoms)
            assert self.calculation is not None
            self.converge()
        elif changes:
            self.move_atoms(atoms)
            self.converge()

        if prop not in self.calculation.results:
            if prop == 'forces':
                with self.timer('Forces'):
                    self.calculation.forces()
            elif prop == 'stress':
                with self.timer('Stress'):
                    self.calculation.stress()
            elif prop == 'dipole':
                self.calculation.dipole()
            else:
                raise ValueError('Unknown property:', prop)

        return self.calculation.results[prop] * units[prop]

    def create_new_calculation(self, atoms: Atoms) -> None:
        with self.timer('Init'):
            self.calculation = DFTCalculation.from_parameters(
                atoms, self.params, self.log)
        self.atoms = atoms.copy()

    def move_atoms(self, atoms):
        with self.timer('Move'):
            self.calculation = self.calculation.move_atoms(atoms)
        self.atoms = atoms.copy()

    def converge(self):
        """Iterate to self-consistent solution.

        Will also calculate "cheap" properties: energy, magnetic moments
        and dipole moment.
        """
        with self.timer('SCF'):
            self.calculation.converge()

        # Calculate all the cheap things:
        self.calculation.energies()
        self.calculation.dipole()
        self.calculation.magmoms()

        self.calculation.write_converged()

    def __del__(self):
        try:
            self.log('---')
            self.timer.write(self.log)
            mib = maxrss() / 1024**2
            self.log(f'\nMax RSS: {mib:.3f}  # MiB')
        except NameError:
            pass

    def get_potential_energy(self,
                             atoms: Atoms,
                             force_consistent: bool = False) -> float:
        return self.calculate_property(atoms,
                                       'free_energy' if force_consistent else
                                       'energy')

    def get_forces(self, atoms: Atoms) -> Array2D:
        return self.calculate_property(atoms, 'forces')

    def get_stress(self, atoms: Atoms) -> Array1D:
        return self.calculate_property(atoms, 'stress')

    def get_dipole_moment(self, atoms: Atoms) -> Array1D:
        return self.calculate_property(atoms, 'dipole')

    def get_magnetic_moment(self, atoms: Atoms) -> float:
        return self.calculate_property(atoms, 'magmom')

    def get_magnetic_moments(self, atoms: Atoms) -> Array1D:
        return self.calculate_property(atoms, 'magmoms')

    def write(self, filename, mode=''):
        """Write calculator object to a file.

        Parameters
        ----------
        filename:
            File to be written
        mode:
            Write mode. Use ``mode='all'``
            to include wave functions in the file.
        """
        self.log(f'# Writing to {filename} (mode={mode!r})\n')

        write_gpw(filename, self.atoms, self.params,
                  self.calculation, skip_wfs=mode != 'all')

    # Old API:

    def get_pseudo_wave_function(self, band, kpt=0, spin=0) -> Array3D:
        state = self.calculation.state
        wfs = state.ibzwfs.get_wfs(spin=spin, kpt=kpt, n1=band, n2=band + 1)
        basis = getattr(self.calculation.scf_loop.hamiltonian, 'basis', None)
        grid = state.density.nt_sR.desc
        wfs = wfs.to_uniform_grid_wave_functions(grid, basis)
        psit_R = wfs.psit_nX[0]
        if not psit_R.desc.pbc.all():
            psit_R = psit_R.to_pbc_grid()
        return psit_R.data * Bohr**-1.5

    def get_atoms(self):
        atoms = self.atoms.copy()
        atoms.calc = self
        return atoms

    def get_fermi_level(self) -> float:
        state = self.calculation.state
        fl = state.ibzwfs.fermi_levels * Ha
        assert len(fl) == 1
        return fl[0]

    def get_homo_lumo(self, spin: int = None) -> Array1D:
        state = self.calculation.state
        return state.ibzwfs.get_homo_lumo(spin) * Ha

    def get_number_of_electrons(self):
        state = self.calculation.state
        return state.ibzwfs.nelectrons

    def get_number_of_bands(self):
        state = self.calculation.state
        return state.ibzwfs.nbands

    def get_atomic_electrostatic_potentials(self):
        return self.calculation.electrostatic_potential().atomic_potentials()

    def get_pseudo_density(self, spin=None):
        return self.calculation.densities().pseudo_densities().data

    def get_all_electron_density(self, spin=None, gridrefinement=1):
        assert spin is None
        n_sr = self.calculation.densities().all_electron_densities(
            grid_refinement=gridrefinement)
        return n_sr.data.sum(0)

    def get_eigenvalues(self, kpt=0, spin=0):
        state = self.calculation.state
        return state.ibzwfs.get_eigs_and_occs(k=kpt, s=spin)[0] * Ha

    def get_reference_energy(self):
        return self.calculation.setups.Eref * Ha

    def get_number_of_iterations(self):
        return self.calculation.scf_loop.niter

    def get_bz_k_points(self):
        state = self.calculation.state
        return state.ibzwfs.ibz.bz.kpt_Kc.copy()

    def get_ibz_k_points(self):
        state = self.calculation.state
        return state.ibzwfs.ibz.kpt_kc.copy()

    def calculate(self, atoms):
        self.get_potential_energy(atoms)

    @cached_property
    def wfs(self):
        from gpaw.new.backwards_compatibility import FakeWFS
        return FakeWFS(self.calculation, self.atoms)

    @property
    def density(self):
        from gpaw.new.backwards_compatibility import FakeDensity
        return FakeDensity(self.calculation)

    @property
    def hamiltonian(self):
        from gpaw.new.backwards_compatibility import FakeHamiltonian
        return FakeHamiltonian(self.calculation)

    @property
    def spos_ac(self):
        return self.atoms.get_scaled_positions()

    @property
    def world(self):
        return self.calculation.scf_loop.world

    @property
    def setups(self):
        return self.calculation.setups

    @property
    def initialized(self):
        return self.calculation is not None

    def get_xc_difference(self, xcparams):
        """Calculate non-selfconsistent XC-energy difference."""
        state = self.calculation.state
        xc = XC(xcparams, state.density.ncomponents)
        exct = self.calculation.pot_calc.calculate_non_selfconsistent_exc(
            state.density.nt_sR, xc)
        dexc = 0.0
        for a, D_sii in state.density.D_asii.items():
            setup = self.setups[a]
            dexc += xc.calculate_paw_correction(setup, pack(D_sii))
        return (exct + dexc - state.potential.energies['xc']) * Ha

    def diagonalize_full_hamiltonian(self,
                                     nbands: int = None,
                                     scalapack=None,
                                     expert: bool = None) -> None:
        if expert is not None:
            warnings.warn('Ignoring deprecated "expert" argument')
        state = self.calculation.state
        ibzwfs = diagonalize(state.potential,
                             state.ibzwfs,
                             self.calculation.scf_loop.occ_calc,
                             nbands)
        self.calculation.state = DFTState(ibzwfs,
                                          state.density,
                                          state.potential)
        nbands = ibzwfs.nbands
        self.params.nbands = nbands
        self.params.keys.append('nbands')

    def gs_adapter(self):
        from gpaw.response.groundstate import ResponseGroundStateAdapter
        return ResponseGroundStateAdapter(self)

    def fixed_density(self, **kwargs):
        kwargs = {**dict(self.params.items()), **kwargs}
        params = InputParameters(kwargs)
        txt = params.txt
        world = params.parallel['world']
        log = Logger(txt, world)
        builder = create_builder(self.atoms, params)
        basis_set = builder.create_basis_set()
        state = self.calculation.state
        ibzwfs = builder.create_ibz_wave_functions(basis_set, state.potential)
        ibzwfs.fermi_levels = state.ibzwfs.fermi_levels
        state = DFTState(ibzwfs, state.density, state.potential)
        scf_loop = builder.create_scf_loop()
        scf_loop.update_density_and_potential = False

        calculation = DFTCalculation(
            state,
            builder.setups,
            scf_loop,
            SimpleNamespace(fracpos_ac=self.calculation.fracpos_ac),
            log)

        calculation.converge()

        return ASECalculator(params, log, calculation, self.atoms)

    def initialize(self, atoms):
        self.create_new_calculation(atoms)


def write_header(log, world, params):
    from gpaw.io.logger import write_header as header
    log(f'#  __  _  _\n# | _ |_)|_||  |\n# |__||  | ||/\\| - {__version__}\n')
    header(log, world)
    log('---')
    with log.indent('input parameters:'):
        log(**{k: v for k, v in params.items()})


def compare_atoms(a1: Atoms, a2: Atoms) -> set[str]:
    if len(a1.numbers) != len(a2.numbers) or (a1.numbers != a2.numbers).any():
        return {'numbers'}
    if (a1.pbc != a2.pbc).any():
        return {'pbc'}
    if abs(a1.cell - a2.cell).max() > 0.0:
        return {'cell'}
    if abs(a1.positions - a2.positions).max() > 0.0:
        return {'positions'}
    return set()



# web-page: lr2_with_08.00eV.out
# web-page: spectrum_with_08.00eV.dat
# web-page: transitions_with_08.00eV.dat
# Start
from gpaw.mpi import world
from gpaw import GPAW
from gpaw.lrtddft2 import LrTDDFT2
from gpaw.lrtddft2.lr_communicators import LrCommunicators

# Maximum energy difference for Kohn-Sham transitions
# included in the calculation
max_energy_diff = 8.0  # eV

# Set parallelization
dd_size = 2 * 2 * 2
eh_size = world.size // dd_size
assert eh_size * dd_size == world.size
lr_comms = LrCommunicators(world, dd_size, eh_size)

calc = GPAW('unocc.gpw',
            communicator=lr_comms.dd_comm)
lr = LrTDDFT2('lr2', calc,
              fxc='LDA',
              max_energy_diff=max_energy_diff,
              recalculate=None,  # Change this to force recalculation
              lr_communicators=lr_comms,
              txt='lr2_with_%05.2feV.out' % max_energy_diff)

# This is the expensive part triggering the calculation!
lr.calculate()

# Get and write spectrum
spec = lr.get_spectrum('spectrum_with_%05.2feV.dat'
                       % max_energy_diff,
                       min_energy=0,
                       max_energy=10,
                       energy_step=0.01,
                       width=0.1)

# Get and write transitions
trans = lr.get_transitions('transitions_with_%05.2feV.dat'
                           % max_energy_diff,
                           min_energy=0.0,
                           max_energy=10.0)



import sys
import numpy as np
from numpy.linalg.linalg import LinAlgError
from ase.units import kJ
import ase.db
from ase.test.tasks.dcdft import DeltaCodesDFTCollection as Collection
from ase.test.tasks.dcdft import FullEquationOfState as EquationOfState

collection = Collection()

db = sys.argv[1]

c = ase.db.connect(db)


def analyse(c, collection):

    A = []
    for name in collection.names:
        ve = []  # volume, energy pairs
        for d in c.select(name=name):
            try:
                ve.append((abs(np.linalg.det(d.cell)), d.energy))
            except AttributeError:
                ve.append((np.nan, np.nan))

        # sort according to volume
        ves = sorted(ve, key=lambda x: x[0])

        # EOS
        eos = EquationOfState([t[0] for t in ves],
                              [t[1] for t in ves])
        try:
            v, e, B0, B1, R = eos.fit()
        except (ValueError, TypeError, LinAlgError):
            (v, e, B0, B1, R) = (np.nan, np.nan, np.nan, np.nan, np.nan)
        e = e / len(collection[name])
        v = v / len(collection[name])
        B0 = B0 / kJ * 1.0e24  # GPa
        A.append((e, v, B0, B1, R))
    return np.array(A).T


E, V, B0, B1, R = analyse(c, collection)

with open(db + '_raw.txt', 'w') as fd:
    for name, e, v, b0, b1, r, in zip(collection.names, E, V, B0, B1, R):
        if not np.isnan(e):
            print('%2s %8.4f %8.4f %8.4f' % (name, v, b0, b1), file=fd)
with open(db + '_raw.csv', 'w') as fd:
    for name, e, v, b0, b1, r, in zip(collection.names, E, V, B0, B1, R):
        if not np.isnan(e):
            print(f'{name}, {v:8.4f}, {b0:8.4f}, {b1:8.4f}', file=fd)



import numpy as np
from math import pi, cos, sin
from ase import Atom, Atoms
from gpaw import GPAW
from gpaw.poisson import FDPoissonSolver
from gpaw.xas import XAS
from gpaw.test import equal
from gpaw.atom.generator2 import generate
import gpaw.mpi as mpi


def test_corehole_h2o(in_tmp_dir, add_cwd_to_setup_paths):
    # Generate setup for oxygen with half a core-hole:
    gen = generate('O', 8, '2s,s,2p,p,d', [1.2], 1.0, None, 2,
                   core_hole='1s,0.5')
    setup = gen.make_paw_setup('hch1s')
    setup.write_xml()

    a = 5.0
    d = 0.9575
    t = pi / 180 * 104.51
    H2O = Atoms([Atom('O', (0, 0, 0)),
                 Atom('H', (d, 0, 0)),
                 Atom('H', (d * cos(t), d * sin(t), 0))],
                cell=(a, a, a), pbc=False)
    H2O.center()
    calc = GPAW(nbands=10, h=0.2, setups={'O': 'hch1s'},
                experimental={'niter_fixdensity': 2},
                poissonsolver=FDPoissonSolver(use_charge_center=True))
    H2O.calc = calc
    _ = H2O.get_potential_energy()

    if mpi.size == 1:
        xas = XAS(calc)
        x, y = xas.get_spectra()
        e1_n = xas.eps_n
        de1 = e1_n[1] - e1_n[0]

    calc.write('h2o-xas.gpw')

    if mpi.size == 1:
        calc = GPAW('h2o-xas.gpw', txt=None,
                    poissonsolver=FDPoissonSolver(use_charge_center=True))
        calc.initialize()
        xas = XAS(calc)
        x, y = xas.get_spectra()
        e2_n = xas.eps_n
        w_n = np.sum(xas.sigma_cn.real**2, axis=0)
        de2 = e2_n[1] - e2_n[0]

        equal(de2, 2.064, 0.005)
        equal(w_n[1] / w_n[0], 2.22, 0.01)

        assert de1 == de2

    if 0:
        import matplotlib.pyplot as plt
        plt.plot(x, y[0])
        plt.plot(x, sum(y))
        plt.show()



# web-page: h.png
def workflow():
    from myqueue.workflow import run
    with run(script='h.py'):
        run(function=plot)


def plot():
    import numpy as np
    import pylab as plt
    from ase.io import read
    with open('h.py') as fd:
        code = fd.read().replace('ae', 'paw')
    exec(code)
    ae = np.array([h.get_potential_energy() for h in read('H.ae.txt@:')])
    paw = np.array([h.get_potential_energy() for h in read('H.paw.txt@:')])
    ecut = range(200, 901, 100)
    plt.figure(figsize=(6, 4))
    plt.plot(ecut, ae[:-1] - ae[-1], label='ae')
    plt.plot(ecut, paw[:-1] - paw[-1], label='paw')
    plt.legend(loc='best')
    plt.xlabel('ecut [eV]')
    plt.ylabel('E(ecut)-E(1000 eV)')
    plt.savefig('h.png')



# web-page: nl_freq_grid.png
import matplotlib.pyplot as plt
import numpy as np
from ase.units import Ha

from gpaw.response.frequencies import NonLinearFrequencyDescriptor

omegamax = 50.0
domega0 = 0.2

plt.figure(figsize=(5, 5))
for omega2 in [2.5, 5, 10, 15, 20, np.inf]:
    x = NonLinearFrequencyDescriptor(
        domega0 / Ha, omega2 / Ha, omegamax / Ha).omega_w * Ha
    y = range(len(x))
    if omega2 == np.inf:
        label = '$\\omega_2 = \\infty$'
    else:
        label = '$\\omega_2 = %.1f\\, \\mathrm{eV}$' % omega2
    plt.plot(x, y, '.', label=label)
plt.ylabel('Freq. no')
plt.xlabel(r'$\omega\, [\mathrm{eV}]$')
plt.axis(xmin=0, xmax=30, ymin=0, ymax=200)
plt.title(r'$\Delta\omega_0 = 0.2\, \mathrm{eV}$')
plt.legend(loc=2)
plt.savefig('nl_freq_grid.png', bbox_inches='tight')



# %%
# teacher
import ase.visualize as viz
viz.view = lambda atoms, repeat=None: None

# %%
"""
# Curie temperature of CrI$_3$

In 2017, ferromagnetic order was observed in a monolayer of CrI$_3$ below 45 K (*Nature* **546** 270 (2017)). It comprises the first demonstration of magnetic order in a 2D material and has received a lot of attention due to the peculiar properties of magnetism in 2D. The physics of magnetic order in 2D is rather different than in 3D and in order to understand what is going on we will need to introduce a bit of theory. But before we get to that let us get started with the calculations.
"""

# %%
"""
## DFT calculation - finding the atomic structure of CrI$_3$

We start by setting up the atomic structure of a CrI$_3$ monolayer and optimizing the atomic positions and unit cell. There are two formula units in the minimal unit cell and only the Cr atoms bear significant magnetic moments. A spin-polarized calculation is initiated by specifying the initial magnetic moments of the all the atoms in units of $\mu_B$.

1. What do you expect the ionic value for the magnetic moment of the Cr atoms to be? (Hint: Use Hund's rule. The electronic configuration of a Cr atom is [Ar]3d$^5$4s$^1$ and each of the iodine atoms will steal one electron). Use your answer to fill in the spin state `S` in the cell below.

Try to understand the individual lines in the input cell below and run it. The calculation will open the ase gui to show the initial atomic structure. You may, for example, try to repeat the structure (under view) to get a better feeling for the material. Then look at the text output below the cell and answer the following questions:

2.  How many electrons are used in the calculation? Which valence states of Cr and I are included?
3.  What is the total magnetic moment after the first DFT calculation and on which atoms are the magnetic moments located?
4.  What is the number of irreducible k-points (k-points not related by symmetry) in the calculation?
5.  What is the maximum force on the atoms after the first DFT calculation? Does it become smaller after subsequent calculations?

In order to get the script running fast, we have set a few of the computational parameters to values which are expected to produce a somewhat inaccurate result. Can you identify what parameters one would need to converge or modify in order to produce more accurate results?

Leave the script running and continue with the theory section below.
"""

# %%
# %load summerschool.py
from gpaw import GPAW, PW
from ase import Atoms
from ase.visualize import view
from ase.optimize import BFGS
from ase.constraints import UnitCellFilter

a = 7.0
c = 10.0
S = 3 / 2  # student: S = ???
m = S * 2                              # Magnetic moment in Bohr magnetons
cell = [[a, 0, 0],
        [-0.5 * a, 3**0.5 / 2 * a, 0],
        [0, 0, c]]                     # Unit cell in \AA
scaled_pos = [[2 / 3., 1 / 3., 0.5],
              [1 / 3., 2 / 3., 0.5],
              [0.6, 1.0, 0.35],
              [0.4, 0.4, 0.35],
              [0.0, 0.6, 0.35],
              [0.4, 1.0, 0.65],
              [0.6, 0.6, 0.65],
              [0.0, 0.4, 0.65]]         # Positions in unit cell vectors
a = Atoms('Cr2I6', cell=cell, scaled_positions=scaled_pos, pbc=True)
a.set_initial_magnetic_moments([m, m, 0, 0, 0, 0, 0, 0])
view(a)

Nk = 2
calc = GPAW(mode=PW(200),
                xc='PBE',
                convergence={'density': 0.001, 'eigenstates': 0.1},
                kpts=(Nk, Nk, 1))
a.calc = calc

uf = UnitCellFilter(a, mask=[1, 1, 0, 0, 0, 1])
opt = BFGS(uf)
opt.run(fmax=0.2)
calc.write('CrI3_relaxed.gpw')

# %%
"""
## A bit of theory

### The Heisenberg model

If we want to calculate the Curie temperature of CrI$_3$ it is, in principle, clear what we have to do. We should calculate the magnetization as a function of temperature using standard methods from statistical physics, and record the temperature where the magnetization vanishes. Unfortunately, this requires knowledge of all the excited states of the system, which we do not have access to. In particular, the magnetization will be dominated by collective magnetic exciations, and these are not directly accessible from the Kohn-Sham spectrum produced by our DFT calculations.

Instead we will consider the Heisenberg Hamiltonian, which captures the basic physics of typical spin systems. It is given by

$$H = -\frac{1}{2}\sum_{i,j}J_{ij}\mathbf{S}_i\cdot \mathbf{S}_j,$$

where $\mathbf{S}_i$ denotes the spin operator at site $i$ in units of $\hbar$ and $J_{ij}$ are magnetic exchange coupling constants. If we want to model a real material with the Heisenberg model we then need to identify a set of magnetic sites and calculate the exchange coupling constants $J_{ij}$.

1.   What are the magnetic sites of CrI$_3$?
2.   How many nearest neighbors does each magnetic site have?
3.   What are the possible values of $S_i^z$ for a magnetic site in CrI$_3$?
4.   What is the unit of $J_{ij}$?

In the following, we will assume that the physics is dominated by neareast neighbor interactions such that $J_{ij}\equiv J$ if atoms $i$ and $j$ are nearest neighbors and zero otherwise. In 3D systems a reasonable estimate of the Curie temperature can be obtained from mean-field theory as

$$T_c^{\mathrm{MF}}=\frac{NJS(S+1)}{3k_B},$$

where $k_B$ is Boltzmann's constant, $N$ is the number of nearest neighbors, and $S$ is the maximum value of $S^z_i$.
"""

# %%
"""
## DFT calculation of $J$

We now want to make a first principles calculation of the nearest neighbor exchange coupling constant $J$. Since the exchange coupling parametrizes the energy difference between aligned and anti-aligned spin configurations, we can obtain $J$ by considering the energy difference between a ferromagnetic and an antiferromagnetic calculation. Note that both can be obtained as collinear DFT ground states subject to different spin constraints. For the CrI$_3$ system, $J$ can calculated as

$$J=\frac{E_{\mathrm{AFM}}-E_{\mathrm{FM}}}{3S^2},$$

where $E_{\mathrm{FM}}$ and $E_{\mathrm{AFM}}$ are the energies *per magnetic atom* of the ferromagnetic and antiferromagnetic configurations respectively.

1.   Derive the expression for $J$ from the classical Heisenberg model yourself. In particular, how does the factor of 3 arise?

We have compiled a database of various 2D materials at https://cmrdb.fysik.dtu.dk/c2db/, which are relaxed with the PBE functional. We will therefore refrain from doing a full coverged geometry optimization and simply download the optimized structure from the database. Search the database for CrI$_3$ (it appears as Cr$_2$I$_6$ on the webpage). If you like, you can take a look at various properties of the material like band structure and stability. Download the `.xyz` file and save it as `CrI3.xyz`. You can also download the structure file directly from the summer school project site https://wiki.fysik.dtu.dk/gpaw/summerschools/summerschool22/magnetism/magnetism.html. With the input structure downloaded, run the cell below to obtain a `.gpw` file containing a converged ferromagnetic calculation. The calculation will take about 30 minutes. To speed up the process, you can copy the cell contents to a python script and submit it as a batch job to the DTU computers with multiple CPU cores. To do so, follow the instructions [here](https://wiki.fysik.dtu.dk/gpaw/summerschools/summerschool22/submitting.html). If the relaxation in the cell above did not finish you may kill it. It is not crucial to complete in order to proceed with the rest of the exercise. Continue with the theory below while you wait for the calculations to finish.
"""

# %%
from ase.io import read
from gpaw import GPAW, PW

a = read('CrI3.xyz')
a.set_initial_magnetic_moments([m, m, 0, 0, 0, 0, 0, 0])

Nk = 4
calc = GPAW(mode=PW(400), xc='PBE', kpts=(Nk, Nk, 1), symmetry='off')
a.calc = calc
a.get_potential_energy()
calc.write('CrI3_fm.gpw')


# %%
"""
Now run the cell again but change the initial magnetic configuration such that it becomes antiferromagnetic (remember to change the name of the `.gpw` file so you do not overwrite the `CrI3_fm.gpw` file). In practise, we do not need to constrain the antiferromagnetic calculation because it comprises a local minimum, but check the magnetic moments in the output to verify that we do indeed end up with an antiferromagnetic state!
"""

# %%
# teacher:
a = read('CrI3.xyz')
a.set_initial_magnetic_moments([m, -m, 0, 0, 0, 0, 0, 0])

Nk = 4
calc = GPAW(mode=PW(400), xc='PBE', kpts=(Nk, Nk, 1))
a.calc = calc
a.get_potential_energy()
calc.write('CrI3_afm.gpw')

# %%
"""
Finally, we can calculate $J$ and $T_c^{\mathrm{MF}}$ by extracting the *ab initio* energy difference from the `.gpw` files. Fill in the formulas for `J` and `T_c` below and evaluate the cell.
"""

# %%
calc_fm = GPAW('CrI3_fm.gpw', txt=None)      # Ferromagnetic calculation
calc_afm = GPAW('CrI3_afm.gpw', txt=None)    # Anti-ferromagnetic calculation

N = 3
E_fm = calc_fm.get_potential_energy() / 2    # Energy per magnetic atom
E_afm = calc_afm.get_potential_energy() / 2  # Energy per magnetic atom
dE = E_afm - E_fm

J = dE / S**2 / 3  # student: J = ???
print(f'J = {J * 1000:1.3f} meV')

from ase.units import kB                     # Boltzmann's constant in eV/K
T_c = N * J * S * (S + 1) / 3 / kB  # student: T_c = ???
print(f'T_c(MF) = {T_c:1.1f} K')

# %%
"""
## More theory

### The Mermin-Wagner theorem
Completing the previous calculations, you should have obtained a value of $T_c$, which is on the order of 100 K. This is much larger than the experimental value.  However in 2D materials mean-field theory fails miserably and the results cannot be trusted. In fact, at finite temperatures the Heisenberg model stated above does not exhibit magnetic order in two dimensions. The reason is that entropy is dominant over enthalpy, such that the free energy is always minimized by disordered configurations at finite temperatures. This is summarized by the Mermin-Wagner theorem, which states that:

*Continuous symmetries cannot be spontaneously broken at finite temperature for systems with short range interactions in dimensions $d\le2$*.

The Heisenberg model above has a continuous rotational symmetry in the spin degrees of freedom and magnetic order is obtained by choosing a certain direction for all the spins. This means that the spin rotation symmetry is spontaneously broken in the magnetically ordered state. However, the direction of magnetization is arbitrary and can still be rotated without any energy cost, as long as the spins remain aligned with respect to each other.

In the Heisenberg model it is straightforward to calculate the collective magnetic excitations of the system, yielding a quadratic spin wave dispersion for ferromagnetic systems

$$\varepsilon(q)=Dq^2$$

in the limit of $q\rightarrow 0$. The spin wave excitations are bosons lowering the total spin along the magnetized direction by a single unit. Hence, the magnetization at finite temperatures $T$ can be calculated from

$$M=M_0 - \int_0^\infty\frac{g(\epsilon)d\varepsilon}{e^{\varepsilon/k_B T} - 1}$$

where $M_0$ is the ground state magnetization and $g(\varepsilon)$ is the density of states for the spin waves.

1.   Calculate the spin wave density of states $g(\varepsilon)$ as a function of dimensionality $d$. (Answer: $g(\varepsilon)=a_d D^{-d/2} \varepsilon^{(d-2)/2}$, where $D$ is the spin wave stiffness defined above and $a_d$ is a geometric constant depending on the dimension $d$)
2.   Compute the bosonic occupation and show that it diverges for $d\le2$. (Hint: You can convince yourself that it does by Taylor expanding the $\varepsilon\rightarrow 0$ limit of the integral)

The divergence of the integral for $d\le2$ shows that the ferromagnetic ground state is thermodynamically unstable and comprises an example of the Mermin-Wagner theorem: In $d\le2$, the free energy is always dominated by entropy at finite temperatures and magnetic order cannot be maintained if a material preserves the rotational symmetry of the spin degrees of freedom. The instability in two dimensions is closely related to the vanishing energy of magnetic excitations in the limit of $q\rightarrow0$, which in turn is a consequence of the rotational symmetry.

### Magnetic anisotropy
Thanks to the Mermin-Wagner theorem, magnetic order is only possible in two dimensions if the rotational symmetry of the spins is *explicitly broken*. That is, there must be an additional term in the Hamiltonian that breaks the symmetry. Such a term can be provided by spin-orbit coupling which couples the spin to the lattice through the electronic orbitals.
We assume that CrI$_3$ is isotropic in the plane of the monolayer and introduce an anisotropy term in the Heisenberg Hamiltonian of the form

$$H_{\mathrm{ani}}=A\sum_i(S_i^z)^2,$$

where we have chosen the $z$-direction to be orthogonal to the plane.

3.   Describe the physics of this term in the cases of $A<0$ and $A>0$. Does the term fully break rotational symmetry of the ground state in both cases?


## Magnetic anisotropy from DFT

In the cell below, the magnetic anisotropy is calculated for the ferromagnetic ground state. The function `calculate_band_energy()` will return the sum of Kohn-Sham eigenvalues for the occupied states. This energy will depend on the direction of the spins, which is specified by the polar and azimuthal angles $\theta$ and $\varphi$ respectively.

1.   What is the sign of $A$ in the Hamiltonian above? (Fill in the formula for `A` and evaluate the cell)
2.   Does spin-orbit coupling break the rotational symmetry of the ground state?

"""

# %%
from gpaw.spinorbit import soc_eigenstates

e_x = soc_eigenstates(calc_fm, theta=90, phi=0).calculate_band_energy() / 2
e_y = soc_eigenstates(calc_fm, theta=90, phi=90).calculate_band_energy() / 2
e_z = soc_eigenstates(calc_fm, theta=0, phi=0).calculate_band_energy() / 2
de_zx = e_z - e_x
de_zy = e_z - e_y
print(f'dE_zx = {de_zx * 1000:1.3f} meV')
print(f'dE_zy = {de_zy * 1000:1.3f} meV')
A = (de_zx + de_zy) / 2 / S**2 # student: A = ???
print(f'A = {A * 1000:1.3f} meV')


# %%
"""
We can also plot the total energy of the ground state as a function of the polar angle $\theta$.

3.   Run the cell below and inspect the plot. Does it look like you would expect?

"""

# %%
import matplotlib.pyplot as plt
import numpy as np

thetas = np.linspace(0, 180, 13)

e_n = []
for theta in thetas:
    soc = soc_eigenstates(calc_fm, theta=theta, phi=0)
    e_n.append(soc.calculate_band_energy() / 2)

e_n = np.array(e_n) - e_n[0]

plt.figure()
plt.plot(thetas, e_n * 1000, 'o-')
plt.xlabel(r'$\theta$', size=18)
plt.ylabel('E [meV]', size=18)


# %%
"""
Now that we have calculated the anisotropy constant $A$, we are finally in a position to improve our estimate of the Curie temperature of CrI$_3$. But how do we get the critical temperature if we cannot apply mean-field theory? One way is to perform Monte-Carlo simulations of the classical Heisenberg model as a function of temperature and find the point where the total magnetization vanishes. The results of such simulations are well approximated by the expression [[2D Mater. 6 (2019) 015028]](https://iopscience.iop.org/article/10.1088/2053-1583/aaf06d)

$$T_c=T_c^{\mathrm{Ising}}\tanh^{1/4}\Big[\frac{6}{N}\log\Big(1-0.033\frac{A}{J}\Big)\Big],$$

where $T_c^{\mathrm{Ising}}=1.52\cdot S^2J/k_B$ is the critical temperature of the ferromagnetic Ising model on a honeycomb lattice. 

4.   Fill in the Monte-Carlo formula for `T_c` in the cell below and calculate the Curie temperature using the values of $A$ and $J$ found above.

"""

# %%
from numpy import tanh, log

T0 = 1.52
T_c = T0 * S**2 * J / kB * (tanh(6 / N * log(1 - 0.033 * A / J)))**(0.25)  # student: T_c = ???
print(f'T_c = {T_c:.1f} K')


# %%
"""
The result for $T_c$ should be in reasonable agreement with the experimental value of 45 K (expect to obtain a value around 30 K). Of course one should carefully check the convergence of all calculations in the present notebook. In fact a converged calculation yields $J = 3.1$ meV and $A=-0.38$ meV, which results in $T_c = 37$ K.
"""



from ase.parallel import paropen
from ase.units import Hartree
from gpaw.xc.rpa import RPACorrelation
from gpaw.xc.fxc import FXCCorrelation

fxc = FXCCorrelation('diamond.ralda.pbe_wfcs.gpw', xc='rAPBE',
                     txt='diamond.ralda_02_rapbe.txt')
E_i = fxc.calculate(ecut=400)

f = paropen('diamond.ralda.rapbe.dat', 'w')
for ecut, E in zip(fxc.ecut_i, E_i):
    print(ecut * Hartree, E, file=f)
f.close()

rpa = RPACorrelation('diamond.ralda.pbe_wfcs.gpw',
                     txt='diamond.ralda_02_rpa.txt')
E_i = rpa.calculate(ecut=400)

f = paropen('diamond.ralda.rpa.dat', 'w')
for ecut, E in zip(rpa.ecut_i, E_i):
    print(ecut * Hartree, E, file=f)
f.close()



import numpy as np
from gpaw.response.df import DielectricFunction

df = DielectricFunction('gs_MoS2.gpw',
                        frequencies=[0.5],
                        txt='eps_GG.txt',
                        hilbert=False,
                        ecut=50,
                        nbands=50)
df_t = DielectricFunction('gs_MoS2.gpw',
                          frequencies=[0.5],
                          txt='eps_GG.txt',
                          hilbert=False,
                          ecut=50,
                          nbands=50,
                          truncation='2D')

for iq in range(22):
    pd, eps_wGG, chi_wGG = df.get_dielectric_matrix(q_c=[iq / 42, iq / 42, 0],
                                                    symmetric=False,
                                                    calculate_chi=True)
    Gvec_Gv = pd.get_reciprocal_vectors(add_q=False)
    epsinv_GG = np.linalg.inv(eps_wGG[0])
    z0 = pd.gd.cell_cv[2, 2] / 2  # Center of layer

    eps_t_wGG = df_t.get_dielectric_matrix(q_c=[iq / 42, iq / 42, 0],
                                           symmetric=False)
    epsinv_t_GG = np.linalg.inv(eps_t_wGG[0])

    epsinv = 0.0
    epsinv_t = 0.0

    for iG in range(len(Gvec_Gv)):
        if np.allclose(Gvec_Gv[iG, :2], 0.0):
            Gz = Gvec_Gv[iG, 2]
            epsinv += np.exp(1.0j * Gz * z0) * epsinv_GG[iG, 0]
            epsinv_t += np.exp(1.0j * Gz * z0) * epsinv_t_GG[iG, 0]

    with open('2d_eps.dat', 'a') as f:
        print(iq, (1.0 / epsinv).real, (1.0 / epsinv_t).real, file=f)



scalapack = True

libraries = ['scalapack-lam',
             'blacsCinit-lam',
             'blacsF77init-lam',
             'blacs-lam',
             'lapack']

define_macros += [('GPAW_NO_UNDERSCORE_CBLACS', '1')]
define_macros += [('GPAW_NO_UNDERSCORE_CSCALAPACK', '1')]



from ase import Atoms
from gpaw import GPAW
from gpaw.cdft.cdft import CDFT
from gpaw import Mixer
import numpy as np


def test_cdft_fd_forces(in_tmp_dir):

    sys = Atoms('N2', positions=([0., 0., 0.], [0., 0., 1]))
    sys.center(3)
    sys.set_pbc(False)
    sys.set_initial_magnetic_moments([0.5, 0.5])

    calc_b = GPAW(h=0.2,
                  basis='dzp',
                  charge=1,
                  mode='lcao',
                  xc='PBE',
                  spinpol=True,
                  mixer=Mixer(beta=0.25, nmaxold=3, weight=100.0),
                  txt='N2.txt',
                  convergence={'eigenstates': 1.0e-2, 'density': 1.0e-2,
                               'energy': 1e-2})

    sys.calc = calc_b
    sys.get_potential_energy()

    cdft_b = CDFT(calc=calc_b,
                  atoms=sys,
                  charge_regions=[[0]],
                  charges=[1],
                  charge_coefs=[22],
                  forces='fd',
                  method='L-BFGS-B',
                  txt='N2.cdft',
                  minimizer_options={'gtol': 0.1})

    sys.calc = cdft_b
    sys.get_potential_energy()
    f = sys.get_forces()
    # check that forces in x and y directions are small
    for i in [0, 1]:
        assert (np.isclose(f[0, i], 0.))
        assert (np.isclose(f[1, i], 0.))



"""Test Hirshfeld for spin/no spin consistency."""
from ase import Atom
from gpaw import GPAW
from ase.parallel import parprint
from gpaw.analyse.hirshfeld import HirshfeldPartitioning
from gpaw import FermiDirac
from gpaw.cluster import Cluster
from gpaw.test import equal


def test_vdw_H_Hirshfeld():
    h = 0.25
    box = 3

    atoms = Cluster()
    atoms.append(Atom('H'))
    atoms.minimal_box(box)

    volumes = []
    for spinpol in [False, True]:
        calc = GPAW(h=h,
                    occupations=FermiDirac(0.1, fixmagmom=spinpol),
                    experimental={'niter_fixdensity': 2},
                    spinpol=spinpol)
        calc.calculate(atoms)
        vol = HirshfeldPartitioning(calc).get_effective_volume_ratios()
        volumes.append(vol)
    parprint(volumes)
    equal(volumes[0][0], volumes[1][0], 4e-9)



# web-page: emt-iterations.csv, lcao-time.csv, systems.csv
import ase.db
from ase.optimize.test.analyze import analyze

analyze('results-emt.db', 'emt')
analyze('results-lcao.db', 'lcao')

db1 = ase.db.connect('systems.db')

with open('systems.csv', 'w') as f:
    print('test-name,description', file=f)
    for row in db1.select():
        print(f'{row.name},{row.description}', file=f)



from ase.build import molecule
from ase.optimize import BFGS
from ase.parallel import paropen
from gpaw import GPAW, LCAO
from gpaw.mom import prepare_mom_calculation
from gpaw.directmin.tools import excite
from gpaw.directmin.etdm import ETDM


for spinpol in [True, False]:
    if spinpol:
        tag = 'spinpol'
    else:
        tag = 'spinpaired'

    atoms = molecule('CO')
    atoms.center(vacuum=5)

    calc = GPAW(xc='PBE',
                mode=LCAO(force_complex_dtype=True),
                h=0.2,
                basis='dzp',
                spinpol=spinpol,
                eigensolver='etdm',
                occupations={'name': 'fixed-uniform'},
                mixer={'backend': 'no-mixing'},
                nbands='nao',
                symmetry='off',
                txt='co_' + tag + '.txt')
    atoms.calc = calc

    # Ground-state calculation
    E_gs = atoms.get_potential_energy()

    # Prepare initial guess for complex pi* orbitals by taking
    # linear combination of real pi*x and pi*y orbitals
    lumo = 5  # lumo is pi*x or pi*y orbital
    for kpt in calc.wfs.kpt_u:
        pp = kpt.C_nM[lumo] + 1.0j * kpt.C_nM[lumo + 1]
        pm = kpt.C_nM[lumo] - 1.0j * kpt.C_nM[lumo + 1]
        kpt.C_nM[lumo][:] = pm
        kpt.C_nM[lumo + 1][:] = pp

    calc.set(eigensolver=ETDM(searchdir_algo={'name': 'l-sr1p'},
                              linesearch_algo={'name': 'max-step'},
                              need_init_orbs=False))

    # Occupation numbers for sigma->pi* excited state:
    # Remove one electron from homo (sigma) and add one electron to lumo (pi*)
    f = excite(calc, 0, 0, spin=(0, 0))
    if not spinpol:
        f[0] /= 2

    # Prepare excited-state DO-MOM calculation
    prepare_mom_calculation(calc, atoms, f)

    opt = BFGS(atoms, logfile='co_' + tag + '.log', maxstep=0.05)
    opt.run(fmax=0.05)

    d = atoms.get_distance(0, 1)

    with paropen('co_' + tag + '.log', 'a') as fd:
        print(f'Optimized CO bond length sigma->pi* state: {d:.2f} Å', file=fd)
        # https://doi.org/10.1007/978-1-4757-0961-2
        print('Experimental CO bond length sigma->pi* state: 1.24 Å', file=fd)



def workflow():
    from myqueue.workflow import run
    td_jobs = [run(script='lcao_linres_NaD.py', tmax='1h'),
               run(script='lcao_linres_NaD2chain.py', tmax='1h')]

    # Plot spectra
    run(script='plot_spectra.py', deps=td_jobs)
    run(script='plot_dipoles.py', deps=td_jobs)
